from __future__ import annotations
import json
import typing as t

from .base import ApiData, ApiValidationFailedError
from ..http import (
    ContentType,
    DEFAULT_CONTENT_TYPE_JSON,
    MediaTypes,
)
from ..util.deco import cached_property, class_property
from ..util.typing import get_args, get_origin


NoneType = type(None)
TYPES_ARGS = (int, float, str, bool, dict, list, NoneType)
TYPES_ARGS_SET = set(TYPES_ARGS)
TYPES_ORIGIN = (list, t.Union)


def _check_annotation(typ: t.Type) -> None:
    origin = get_origin(typ)
    if origin is None:
        if issubclass(typ, TYPES_ARGS):
            return
        elif issubclass(typ, JsonApiData):
            _has_valid_annotations(typ)
        else:
            raise TypeError(
                f"{typ.__name__} is an unacceptable as a type "
                "of attributes of JsonApiData."
            )
    elif origin == list:
        args = get_args(typ)
        if not len(args):
            raise TypeError(
                "List with no arguments is not acceptable. "
                "Specify an argument of type to list."
            )
        _check_annotation(args[0])
    elif origin == t.Union:
        # Only Optional is admitted.
        args = get_args(typ)
        if not (len(args) == 2 and args[1] == NoneType):
            raise TypeError(
                "Attritbutes with multiple types are not acceptable. "
                "Only 'Optional' can be used as type hints."
            )
        _check_annotation(args[0])
    else:
        raise TypeError(
            f"{typ.__name__} is not acceptable as a type of"
            " attributes of JsonApiData."
            )


def _has_valid_annotations(apiclass: t.Type[JsonApiData]) -> None:
    for typ in t.get_type_hints(apiclass).values():
        _check_annotation(typ)


def _validate_obj(obj: t.Any, objtype: t.Type) -> bool:
    origin = get_origin(objtype)
    if origin is None:
        if objtype in TYPES_ARGS_SET:
            return isinstance(obj, objtype)
        if issubclass(objtype, JsonApiData):
            return isinstance(obj, (dict, objtype))
        return False
    elif origin == list:
        if len(obj):
            item = obj[0]
            inner_type = get_args(objtype)[0]
            return _validate_obj(item, inner_type)
        return True
    elif origin == t.Union:
        is_good = False
        for arg in get_args(objtype):
            is_good = is_good or _validate_obj(obj, arg)
        return is_good
    else:
        return False


def _build_json_api(
    apiclass: t.Type[JsonApiData],
    data: t.Dict[str, t.Any]
) -> JsonApiData:
    instance = apiclass.__new__(apiclass)
    annotations = t.get_type_hints(apiclass)

    # NOTE
    #   Ignore keys of data which is not defined in the apiclass.
    for key_def, type_def in annotations.items():

        # Handle default values.
        if not (key_def in data or hasattr(apiclass, key_def)):
            raise ApiValidationFailedError(
                f"Raw data has no key named '{key_def}'."
            )
        if key_def in data:
            val = data.get(key_def)
        else:
            val = getattr(apiclass, key_def)

        # Validate types of values.
        if not _validate_obj(val, type_def):
            raise ApiValidationFailedError(
                "Invalid type was detected in received json data. "
                f"Expected: {type_def.__name__}; "
                f"Detected: {val.__class__.__name__}."
            )

        # Set values of attributes to an instance of the apiclass.
        # Handle only the case in which the origin is t.Optional.
        origin = get_origin(type_def)
        if origin == t.Union:
            type_def = get_args(type_def)[0]
        _set_non_optional_value(instance, type_def, key_def, val)

    return instance


def _set_non_optional_value(
    instance: JsonApiData,
    type_def: t.Any,
    key_def: str,
    val: t.Any,
) -> None:
    origin = get_origin(type_def)
    if origin is None:
        if issubclass(type_def, JsonApiData) and isinstance(val, dict):
            setattr(instance, key_def, type_def(**val))
        else:
            setattr(instance, key_def, val)
    elif origin == list:
        setattr(instance, key_def, _build_list(type_def, val))
    else:
        raise ApiValidationFailedError(
            "Something was wrong. Contact the maintainer of the framework."
        )


def _build_list(
    type_def: t.Type[list],
    val: list
) -> t.List[t.Any]:
    if not len(val):
        return val
    if not (get_origin(type_def) == list and len(get_args(type_def))):
        raise ValueError("'type_def' must be list with an argument.")

    # Check type of inner value
    inner_type = get_args(type_def)[0]
    inner_origin = get_origin(inner_type)
    if inner_origin is None:
        if issubclass(inner_type, JsonApiData):
            return [
                inner_type(**item) if isinstance(item, dict) else item
                for item in val
            ]
        return val
    elif inner_origin == list:
        return [_build_list(inner_type, item) for item in val]
    else:
        raise ApiValidationFailedError(
            f"{inner_type.__name__} is not acceptable as a type of "
            "attributes of JsonApiData."
        )


def _extract_dict(api: JsonApiData) -> t.Dict[str, t.Any]:
    cls = type(api)
    res = {}

    for key, type_def in t.get_type_hints(cls).items():
        if hasattr(api, key):
            val = getattr(api, key)
            if isinstance(val, JsonApiData):
                val = _extract_dict(val)
            elif isinstance(val, list):
                inner_type = get_args(type_def)[0]
                if issubclass(inner_type, JsonApiData):
                    val = [_extract_dict(v) for v in val]
        else:
            val = getattr(cls, key)
        res[key] = val

    return res


class JsonApiData(ApiData):
    """API data with JSON format.

    This class can be used to describe data with JSON format. This class
    should be inheritted and its several class-attributes should be
    defined in the subclass. Developers must define type hints of the
    class-attribtues, which are used to validate if raw data has format
    the type hints define.

    Examples:
        - Defining subclass of this class
        ```python
        class User(JsonApiData):
            name: str
            email: str
            age: int

        class MockApiData(JsonApiData):
            users: List[User]
        ```

        - Validating received data
        ```python
        class MockEndpoint(Endpoint):

            @data_format(input=MockApiData, output=None)
            def do_GET(self, rec_body: MockApiData) -> None:
                # Do something...

                # Example
                for user in rec_body.users:
                    print(f"user name : {user.name}")
        ```
    """

    def __init_subclass__(cls) -> None:
        _has_valid_annotations(cls)

    # NOTE
    #   DO NOT override the method. This class should be used only
    #   by defining attributes and their types.
    def __init__(self, **data: t.Any) -> None:
        mapped = _build_json_api(self.__class__, data)
        self.__dict__.update(mapped.__dict__)

    @classmethod
    def __validate__(
        cls,
        raw: bytes,
        content_type: ContentType
    ) -> JsonApiData:
        """
        Args:
            raw: Raw data to be validated.
            content_type: Values of `Content-Type` header.
        """
        super().__init__(raw, content_type)

        if content_type.charset is None:
            content_type.charset = "UTF-8"

        if not cls.verify_content_type(content_type):
            raise ApiValidationFailedError(
                "Media type of 'Content-Type' header was not "
                f"{MediaTypes.json}, but {content_type.media_type}."
            )

        try:
            raw = raw.decode(encoding=content_type.charset)
            data = json.loads(raw)
        except UnicodeDecodeError:
            raise ApiValidationFailedError(
                "Decoding raw data failed. The encoding was expected "
                f"{content_type.charset}, but not corresponded."
            )
        except json.decoder.JSONDecodeError:
            raise ApiValidationFailedError(
                "Decoding raw data failed."
                "The raw data had invalid JSON format."
            )
        return cls(**data)

    def __extract__(self) -> bytes:
        encoding = self.__content_type__.charset
        if encoding is None:
            encoding = "utf-8"
        return json.dumps(self.dict).encode(encoding=encoding)

    @cached_property
    def dict(self) -> t.Dict[str, t.Any]:
        return _extract_dict(self)

    @class_property
    def __content_type__(cls) -> ContentType:
        """Content type with `application/json; charset=UTF-8`.
        """
        return DEFAULT_CONTENT_TYPE_JSON



from __future__ import annotations
from abc import ABCMeta, abstractmethod
import codecs
import inspect
import io
import json
import os
import typing as t
from urllib.parse import parse_qs

from .api.base import ApiData
from .asgi import (
    ASGIHTTPEvents,
    WebSocketAccept_t,
    WebSocketClose_t,
    WebSocketRecvMsg_t,
    WebSocketSendMsg_t,
)
from .http import (
    ContentType,
    DEFAULT_CONTENT_TYPE_PLAIN,
    HTTPMethods,
    HTTPStatus,
    MediaTypes,
)
from .io import BufferedConcatIterator, BufferedFileIterator
from .util.deco import (
    awaitable_property,
    awaitable_cached_property,
    cached_property,
)
from .util.header import make_header


if t.TYPE_CHECKING:
    from .app import (
        AppBase,
        WSGIApp,
        ASGIHTTPApp,
    )

    App_t = t.TypeVar("App_t", bound=AppBase)


__all__ = []


# Base classes for each interfaces  ------------------------------------------

class EndpointBase(metaclass=ABCMeta):
    """Base class of Endpoint to define logic to requests.

    Endpoint is one of the core concept of Bamboo, and this class defines
    its basic behavior. All endpoints must inherit this class.

    Note:
        This class is an abstract class. Consider using its subclasses.
    """

    def __init__(
        self,
        app: App_t,
        flexible_locs: t.Tuple[str, ...],
        *parcel: t.Any
    ) -> None:
        """
        Note:
            DO NOT generate its instance. Its object will be initialized
            by application object.

        Args:
            flexible_locs: Flexible locations requested.
            *parcel: Parcel sent via application object.
        """
        self._app = app
        self._flexible_locs = flexible_locs
        self.setup(*parcel)

    def setup(self, *parcel) -> None:
        """Execute setup of the endpoint object.

        This method will execute at initialization of the object by specifying
        parcel. The parcel is sent with `set_parcel()` method of
        the application object which has included the object as one of
        the its endpoints. This method can be used as a substitute for
        the `__init__` method.

        This method is useful in some cases like below:

        - Making an endpoint class a reusable component
        - Injecting environmental dependencies using something like a
            setting file

        Args:
            *parcel: Parcel sent via application object.

        Examples:
            ```python
            app = WSGIApp()

            @app.route("hello")
            class HelloEndpoint(WSGIEndpoint):

                def setup(self, server_name: str) -> None:
                    self._server_name = server_name

                def do_GET(self) -> None:
                    self.send_body(f"Hello from {self._server_name}".encode())

            if __name__ == "__main__":
                SERVER_NAME = "awesome_server"
                app.set_parcel(HelloEndpoint, SERVER_NAME)

                WSGITestExecutor.debug(app, "", 8000)
            ```
        """
        pass

    @property
    def app(self) -> App_t:
        """Application object handling the endpoint.
        """
        return self._app

    @property
    def flexible_locs(self) -> t.Tuple[str, ...]:
        """Flexible locations extracted from requested URI.
        """
        return self._flexible_locs

    @property
    @abstractmethod
    def http_version(self) -> str:
        """HTTP Version on communication.
        """
        pass

    @property
    @abstractmethod
    def scheme(self) -> str:
        """Scheme of requested URI.
        """
        pass

    @abstractmethod
    def get_client_addr(self) -> t.Tuple[t.Optional[str], t.Optional[int]]:
        """Retrieve client address, pair of its IP address and port.

        Note:
            IP address and port may be None if retrieving the address from
            server application would fail, so it is recommended to confirm
            your using server application's spec.

        Returns:
            Pair of IP and port of client.
        """
        pass

    @abstractmethod
    def get_server_addr(self) -> t.Tuple[t.Optional[str], t.Optional[int]]:
        """Retrive server address, pair of its IP address and port.

        Note:
            IP address and port may be None if retrieving the address from
            server application would fail, so it is recommended to confirm
            your using server application's spec.

        Returns:
            Pair of IP and port of server.
        """
        pass

    @abstractmethod
    def get_host_addr(self) -> t.Tuple[t.Optional[str], t.Optional[int]]:
        """Retrive host name and port from requested headers.

        Returns:
            Pair of host name and port.
        """
        pass

    @abstractmethod
    def get_header(self, name: str) -> t.Optional[str]:
        """Retrive header value from requested headers.

        Args:
            name: Header name.

        Returns:
            Value of header if existing, None otherwise.
        """
        pass

    @property
    @abstractmethod
    def path(self) -> str:
        """Path of requested URI.
        """
        pass

    @cached_property
    @abstractmethod
    def queries(self) -> t.Dict[str, t.List[str]]:
        """Query parameters specified to requested URI.
        """
        pass

    def get_queries(self, name: str) -> t.List[str]:
        """Get values of query parameter.

        Args:
            name: Key name of the parameter

        Returns:
            Value of the parameter. The value of list may have multiple
            items if client specifies the parameter in several times.
        """
        query = self.queries.get(name)
        if query:
            return query
        return []

    @cached_property
    @abstractmethod
    def content_type(self) -> t.Optional[ContentType]:
        """Content type of request body.

        Returns:
            Content type if existing, None otherwise.
        """
        pass


class WSGIEndpointBase(EndpointBase):
    """Base class of endpoints compliant with the WSGI.

    This class implements abstract methods of `EndpointBase` with the WSGI.
    However, this class doesn't implement some methods to structure responses.

    Note:
        DO NOT use this class as the super class of your endpoints. Consider
        to use subclasses of the class like `WSGIEndpoint`.
    """

    def __init__(
        self,
        app: WSGIApp,
        environ: t.Dict[str, t.Any],
        flexible_locs: t.Tuple[str, ...],
        *parcel: t.Any
    ) -> None:
        """
        Args:
            environ: environ variable received from WSGI server.
            flexible_locs: flexible locations requested.
            *parcel: Parcel sent via application object.
        """
        self._environ = environ
        super().__init__(app, flexible_locs, *parcel)

    @property
    def environ(self) -> t.Dict[str, t.Any]:
        """environ variable received from WSGI server.
        """
        return self._environ

    @property
    def wsgi_version(self) -> str:
        """WSGI version number.
        """
        version = self._environ.get("wsgi.version")
        return ".".join(map(str, version))

    @property
    def server_software(self) -> str:
        """Software name of WSGI server.
        """
        return self._environ.get("SERVER_SOFTWARE")

    @property
    def http_version(self) -> str:
        version = self._environ.get("SERVER_PROTOCOL")
        return version.split("/")[1]

    @property
    def scheme(self) -> str:
        return self._environ.get("wsgi.url_scheme")

    def get_client_addr(self) -> t.Tuple[t.Optional[str], t.Optional[int]]:
        client = self._environ.get("REMOTE_ADDR")
        port = self._environ.get("REMOTE_PORT")
        if port:
            port = int(port)
        return (client, port)

    def get_server_addr(self) -> t.Tuple[t.Optional[str], t.Optional[int]]:
        server = self._environ.get("SERVER_NAME")
        port = self._environ.get("SERVER_PORT")
        if port:
            port = int(port)
        return (server, port)

    def get_host_addr(self) -> t.Tuple[t.Optional[str], t.Optional[int]]:
        http_host = self._environ.get("HTTP_HOST")
        if http_host:
            http_host = http_host.split(":")
            if len(http_host) == 1:
                return (http_host[0], None)
            else:
                host, port = http_host
                port = int(port)
                return (host, port)
        return (None, None)

    def get_header(self, name: str) -> t.Optional[str]:
        name = name.upper()
        if name == "CONTENT-TYPE":
            return self.content_type
        if name == "CONTENT-LENGTH":
            return self._environ.get("CONTENT_LENGTH")

        name = "HTTP_" + name.replace("-", "_")
        return self._environ.get(name)

    @property
    def path(self) -> str:
        return self._environ.get("PATH_INFO")

    @cached_property
    def queries(self) -> t.Dict[str, t.List[str]]:
        return parse_qs(self._environ.get("QUERY_STRING"))

    @cached_property
    def content_type(self) -> t.Optional[ContentType]:
        raw = self._environ.get("CONTENT_TYPE")
        if raw:
            return ContentType.parse(raw)
        return None


class ASGIEndpointBase(EndpointBase):
    """Base class of endpoints compliant with the ASGI.

    This class implements abstract methods of `EndpointBase` with the ASGI.
    However, this class doesn't implement some methods to structure responses.

    Note:
        DO NOT use this class as the super class of your endpoints. Consider
        to use subclasses of the class like `ASGIHTTPEndpoint`.
    """

    def __init__(
        self,
        app: App_t,
        scope: t.Dict[str, t.Any],
        flexible_locs: t.Tuple[str, ...],
        *parcel: t.Any
    ) -> None:
        """
        Args:
            scope: scope variable received from ASGI server.
            flexible_locs: Flexible locations requested.
            *parcel: Parcel sent via application object.
        """
        self._scope = scope

        # TODO
        #   Consider not mapping in this method.
        req_headers = scope.get("headers")
        self._req_headers = {}
        if req_headers:
            req_headers = [map(codecs.decode, h) for h in req_headers]
            self._req_headers.update(dict(req_headers))

        super().__init__(app, flexible_locs, *parcel)

    @property
    def scope(self) -> t.Dict[str, t.Any]:
        """scope variable received from ASGI server.
        """
        return self._scope

    @property
    def scope_type(self) -> str:
        """Scope type on ASGI.
        """
        return self._scope.get("type")

    @property
    def asgi_version(self) -> str:
        """ASGI version.
        """
        return self._scope.get("asgi").get("version")

    @property
    def spec_version(self) -> str:
        """Spec version on ASGI.
        """
        return self._scope.get("asgi").get("spec_version")

    @property
    def raw_path(self) -> bytes:
        """The original HTTP path received from client.
        """
        return self._scope.get("raw_path")

    @property
    def root_path(self) -> str:
        """The root path ASGI application is mounted at.
        """
        return self._scope.get("root_path")

    @property
    def http_version(self) -> str:
        return self._scope.get("http_version")

    @property
    def scheme(self) -> str:
        return self._scope.get("scheme")

    def get_client_addr(self) -> t.Tuple[t.Optional[str], t.Optional[str]]:
        client = self._scope.get("client")
        if client:
            return tuple(client)
        return (None, None)

    def get_server_addr(self) -> t.Tuple[t.Optional[str], t.Optional[str]]:
        server = self._scope.get("server")
        if server:
            return tuple(server)
        return (None, None)

    def get_host_addr(self) -> t.Tuple[t.Optional[str], t.Optional[int]]:
        http_host = self.get_header("host")
        if http_host:
            http_host = http_host.split(":")
            if len(http_host) == 1:
                return (http_host[0], None)
            else:
                host, port = http_host
                port = int(port)
                return (host, port)
        return (None, None)

    def get_header(self, name: str) -> t.Optional[str]:
        name = name.lower()
        return self._req_headers.get(name)

    @property
    def headers(self) -> t.Dict[str, str]:
        """Request headers.
        """
        return self._req_headers

    @property
    def path(self) -> str:
        return self._scope.get("path")

    @cached_property
    def queries(self) -> t.Dict[str, t.List[str]]:
        return parse_qs(self._scope.get("query_string").decode())

    @cached_property
    def content_type(self) -> t.Optional[ContentType]:
        raw = self.get_header("Content-Type")
        if raw:
            return ContentType.parse(raw)
        return None


class StatusCodeAlreadySetError(Exception):
    """Raised if response status code has already been set."""
    pass


_AVAILABLE_RES_METHODS = {
    "CONNECT",
    "DELETE",
    "GET",
    "HEAD",
    "OPTIONS",
    "PATCH",
    "POST",
    "PUT",
    "TRACE",
}
_PREFIX_RESPONSE = "do_"
_PREFIX_PRE_RESPONSE = "pre_"


def set_pre_response_method(
    endpoint: t.Type[HTTPMixIn],
    http_method: str,
    callback: t.Callable[[HTTPMixIn], None],
) -> None:
    http_method = http_method.upper()
    if http_method not in HTTPMethods:
        raise ValueError(f"{http_method} is not available as a HTTP method.")

    setattr(endpoint, _PREFIX_PRE_RESPONSE + http_method, callback)
    endpoint._pre_methods[http_method] = callback


def set_response_method(
    endpoint: t.Type[HTTPMixIn],
    http_method: str,
    callback: t.Callable[[HTTPMixIn], None],
) -> None:
    http_method = http_method.upper()
    if http_method not in HTTPMethods:
        raise ValueError(f"{http_method} is not available as a HTTP method.")

    setattr(endpoint, _PREFIX_RESPONSE + http_method, callback)
    endpoint._res_methods[http_method] = callback


class HTTPMixIn(metaclass=ABCMeta):
    """Mixin class for HTTP endpoints.

    This class assumes that endpoint classes inherit this class for HTTP.
    So, this class do not work alone.

    Note:
        DO NOT use this class alone. This class work correctly by inheriting
        it, implementing its abstract methods, and call its `__init__()`
        method in the one of the subclass.
    """
    _pre_methods: t.Dict[str, t.Callable[[HTTPMixIn], None]]
    _res_methods: t.Dict[str, t.Callable[[HTTPMixIn], None]]

    bufsize = 8192

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()

        cls._pre_methods = {}
        cls._res_methods = {}

        # Check if bufsize is positive
        if not (cls.bufsize > 0 and isinstance(cls.bufsize, int)):
            raise ValueError(
                f"{cls.__name__}.bufsize must be positive integer"
            )

        # Check pre & response methods
        for method in _AVAILABLE_RES_METHODS:
            name_pre_method = _PREFIX_PRE_RESPONSE + method
            name_res_method = _PREFIX_RESPONSE + method

            if hasattr(cls, name_pre_method):
                pre_method = getattr(cls, name_pre_method)
                cls._pre_methods[method] = pre_method

            if hasattr(cls, name_res_method):
                res_method = getattr(cls, name_res_method)
                cls._res_methods[method] = res_method


    @classmethod
    def _get_pre_response_method(
        cls,
        method: str
    ) -> t.Optional[t.Callable[[EndpointBase], None]]:
        """Retrieve pre-response method corresponding with given HTTP method.

        Args:
            method: HTTP method

        Returns:
            Pre-response method with given name.
        """
        return cls._pre_methods.get(method, None)

    @classmethod
    def _get_response_method(
        cls,
        method: str
    ) -> t.Optional[t.Callable[[EndpointBase], None]]:
        """Retrieve response methods with given HTTP method.

        Args:
            method: HTTP method

        Returns:
            Callback with given name.
        """
        return cls._res_methods.get(method, None)

    def __init__(self) -> None:
        self._res_status: t.Optional[HTTPStatus] = None
        self._res_headers: t.List[t.Tuple[str, str]] = []
        self._res_body = BufferedConcatIterator(bufsize=self.bufsize)

    @property
    @abstractmethod
    def content_length(self) -> t.Optional[int]:
        """Content length of request body if existing.
        """
        pass

    @property
    @abstractmethod
    def method(self) -> str:
        """HTTP method requested from client.
        """
        pass

    def add_header(
        self,
        name: str,
        value: str,
        **params: str
    ) -> None:
        """Add response header with MIME parameters.

        Args:
            name: Field name of the header.
            value: Value of the field.
            **params: Directives added to the field.
        """
        self._res_headers.append(make_header(name, value, **params))

    def add_headers(self, *headers: t.Tuple[str, str]) -> None:
        """Add response headers at once.

        Note:
            This method would be used as a shortcut to register multiple
            headers. If it requires adding MIME parameters, developers
            can use the 'add_header' method.

        Args:
            **headers: Header's info whose header is the field name.
        """
        for name, val in headers:
            self.add_header(name, val)

    def add_content_type(self, content_type: ContentType) -> None:
        """Add Content-Type header of response.

        Args:
            content_type: Information of Content-Type header.
        """
        self.add_header(*content_type.to_header())

    def add_content_length(self, length: int) -> None:
        """Add Content-Length header of response.

        Args:
            length: Size of response body.
        """
        self.add_header("Content-Length", str(length))

    def add_content_length_body(self, body: bytes) -> None:
        """Add Content-Length header of response by response body.

        Args:
            body: Response body.
        """
        self.add_header("Content-Length", str(len(body)))

    def _set_status_safely(self, status: HTTPStatus) -> None:
        """Check if response status code already exists.

        Raises:
            StatusCodeAlreadySetError: Raised if response status code has
                already been set.
        """
        if self._res_status:
            raise StatusCodeAlreadySetError(
                "Response status code has already been set."
            )
        self._res_status = status

    def send_only_status(self, status: HTTPStatus = HTTPStatus.OK) -> None:
        """Set specified status code to one of response.

        This method can be used if a callback doesn't need to send response
        body.

        Args:
            status: HTTP status of the response.
        """
        self._set_status_safely(status)

    def send_body(
        self,
        body: t.Union[bytes, t.Iterable[bytes]],
        *others: t.Union[bytes, t.Iterable[bytes]],
        content_type: t.Optional[ContentType] = DEFAULT_CONTENT_TYPE_PLAIN,
        status: HTTPStatus = HTTPStatus.OK
    ) -> None:
        """Set given binary to the response body.

        Note:
            If the parameter `content_type` is specified, then
            the `Content-Type` header is to be added.

            `DEFAULT_CONTENT_TYPE_PLAIN` has its MIME type of `text/plain`,
            and the other attributes are `None`. If another value of
            `Content-Type` is needed, then you should generate new
            `ContentType` instance with attributes you want.

        Args:
            body: Binary to be set to the response body.
            content_type: `Content-Type` header to be sent.
            status: HTTP status of the response.

        Raises:
            StatusCodeAlreadySetError: Raised if response status code has
                already been set.
        """
        self._set_status_safely(status)

        bodies = [body]
        bodies.extend(others)

        is_all_bytes = True
        is_empty = False
        for chunk in bodies:
            is_all_bytes &= isinstance(chunk, bytes)
            if is_all_bytes:
                is_empty |= len(chunk) > 0
            self._res_body.append(chunk)

        if content_type:
            self.add_content_type(content_type)

        # Content-Length if avalidable
        if is_all_bytes and not is_empty:
            length = sum(map(len, bodies))
            self.add_content_length(length)

    def send_api(
        self,
        api: ApiData,
        status: HTTPStatus = HTTPStatus.OK,
    ) -> None:
        """Set given api data to the response body.

        Args:
            api: ApiData object to be sent.
            status: HTTP status of the response.

        Raises:
            StatusCodeAlreadySetError: Raised if response status code
                has already been set.
        """
        self.send_body(
            api.__extract__(),
            content_type=api.__content_type__,
            status=status,
        )

    def send_json(
        self,
        body: t.Dict[str, t.Any],
        status: HTTPStatus = HTTPStatus.OK,
        encoding: str = "UTF-8"
    ) -> None:
        """Set given json data to the response body.

        Args:
            body: Json data to be set to the response body.
            status: HTTP status of the response.
            encoding: Encoding of the Json data.

        Raises:
            StatusCodeAlreadySetError: Raised if response status code
                has already been set.
        """
        body = json.dumps(body).encode(encoding=encoding)
        content_type = ContentType(MediaTypes.json, encoding)
        self.send_body(body, content_type=content_type, status=status)

    def send_file(
        self,
        path: str,
        fname: t.Optional[str] = None,
        content_type: str = DEFAULT_CONTENT_TYPE_PLAIN,
        status: HTTPStatus = HTTPStatus.OK
    ) -> None:
        """Set file to be sent as response.

        Args:
            path: File path.
            fname: File name to be sent.
            content_type: Content type of the response.
            status: HTTP status of the response.
        """
        file_iter = BufferedFileIterator(path)
        self.send_body(file_iter, content_type=content_type, status=status)

        length = os.path.getsize(path)
        self.add_header("Content-Length", str(length))
        if fname:
            self.add_header(
                "Content-Disposition",
                "attachment",
                filename=fname
            )


class WSGIEndpoint(WSGIEndpointBase, HTTPMixIn):
    """HTTP endpoint class compliant with the WSGI.

    This class is a complete class of endpoints, communicating on HTTP.
    This class has all attributes of `WSGIEndpointBase` and `HTTPMixIn`,
    and you can define its subclass and use them in your response methods.

    Examples:
        ```python
        app = WSGIApp()

        @app.route("hello")
        class HelloEndpoint(WSGIEndpoint):

            # RECOMMEND to use `data_format` decorator
            def do_GET(self) -> None:
                response = {"greeting": "Hello, Client!"}
                self.send_json(response)

            def do_POST(self) -> None:
                req_body = self.body
                print(req_body)
        ```
    """

    def __init__(
        self,
        app: WSGIApp,
        environ: t.Dict[str, t.Any],
        flexible_locs: t.Tuple[str, ...],
        *parcel: t.Any
    ) -> None:
        """
        Args:
            environ: environ variable received from WSGI server.
            flexible_locs: flexible locations requested.
            *parcel: Parcel sent via application object.
        """
        WSGIEndpointBase.__init__(self, app, environ, flexible_locs, *parcel)
        HTTPMixIn.__init__(self)

    def _recv_body_secure(self) -> bytes:
        """Receive request body securely.

        Returns:
            Request body sent to the endpoint.
        """
        # TODO
        #   Take measures against DoS attack.
        content_length = self.content_length
        if content_length is None:
            content_length = -1

        body = self._environ.get("wsgi.input").read(content_length)
        return body

    @cached_property
    def body(self) -> bytes:
        """Request body received from client.
        """
        return self._recv_body_secure()

    @property
    def content_length(self) -> t.Optional[int]:
        length = self._environ.get("CONTENT_LENGTH")
        if length:
            return int(length)
        return None

    @property
    def method(self) -> str:
        return self._environ.get("REQUEST_METHOD")


class ASGIHTTPEndpoint(ASGIEndpointBase, HTTPMixIn):
    """HTTP endpoint class compliant with the ASGI.

    This class is a complete class of endpoints, communicating on HTTP.
    This class has all attributes of `ASGIEndpointBase` and `HTTPMixIn`,
    and you can define its subclass and use them in your response methods.

    Examples:
        ```python
        app = ASGIHTTPApp()

        @app.route("hello")
        class HelloEndpoint(ASGIHTTPEndpoint):

            # RECOMMEND to use `data_format` decorator
            async def do_GET(self) -> None:
                response = {"greeting": "Hello, Client!"}
                self.send_json(response)

            async def do_POST(self) -> None:
                req_body = async self.body
                print(req_body)
        ```
    """

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()

        # NOTE
        #   All response methods of its subclass must be awaitables.
        for method in HTTPMethods:
            callback = cls._get_response_method(method)
            if callback and not inspect.iscoroutinefunction(callback):
                raise TypeError(
                    f"{cls.__name__}.{callback.__name__} must be an awaitable"
                    ", not a callable."
                )

    def __init__(
        self,
        app: ASGIHTTPApp,
        scope: t.Dict[str, t.Any],
        receive: t.Callable[[], t.Awaitable[t.Dict[str, t.Any]]],
        flexible_locs: t.Tuple[str, ...],
        *parcel
    ) -> None:
        """
        Args:
            scope: scope variable received from ASGI server.
            receive: `receive` method given from ASGI server.
            flexible_locs: Flexible locations requested.
            *parcel: Parcel sent via application object.
        """
        self._receive = receive
        self._is_disconnected = False

        ASGIEndpointBase.__init__(self, app, scope, flexible_locs, *parcel)
        HTTPMixIn.__init__(self)

    @awaitable_cached_property
    async def body(self) -> bytes:
        """Request body received from client.
        """
        buffer = io.BytesIO()
        more_body = True

        while more_body:
            chunk = await self._receive()
            type = chunk.get("type")
            if type == ASGIHTTPEvents.disconnect:
                self._is_disconnected = True
                break

            buffer.write(chunk.get("body", b""))
            more_body = chunk.get("more_body", False)

        buffer.flush()
        data = buffer.getvalue()
        buffer.close()
        return data

    @awaitable_property
    async def is_disconnected(self) -> bool:
        """Whether the connection is closed or not.
        """
        await self.body
        return self._is_disconnected

    @property
    def content_length(self) -> t.Optional[int]:
        length = self.get_header("Content-Length")
        if length:
            return int(length)
        return None

    @property
    def method(self) -> str:
        return self._scope.get("method")


class ASGIWebSocketEndpoint(ASGIEndpointBase):

    @cached_property
    def subprotocols(self) -> t.Tuple[str, ...]:
        return tuple(self._scope.get("subprotocols"))

    async def do_ACCEPT(
        self,
        accept: WebSocketAccept_t,
    ) -> None:
        raise NotImplementedError

    async def do_COMMUNICATE(
        self,
        recvmsg: WebSocketRecvMsg_t,
        sendmsg: WebSocketSendMsg_t,
        close: WebSocketClose_t,
    ) -> None:
        raise NotImplementedError



import json
import typing as t

from .http import (
    AuthSchemes,
    ContentType,
    ContentTypeHolder,
    DEFAULT_CONTENT_TYPE_PLAIN,
    HTTPStatus, MediaTypes,
)
from .io import BufferedConcatIterator
from .util.deco import class_property


__all__ = []


class ErrInfo(Exception, ContentTypeHolder):
    """Base class of all error handlings.

    This class defines the attributes of all classes for error handling.

    Attributes:
        http_status (HTTPStatus) : HTTP status of the error.
        inheritted_headers (Set[str]) : Header names to be inheritted.
    """
    http_status: HTTPStatus = HTTPStatus.BAD_REQUEST
    inheritted_headers: t.Set[str] = set()

    def __init_subclass__(cls) -> None:
        # Make header names lower to ignore the difference of upper and
        # lower charactors.
        cls.inheritted_headers = set(map(str.lower, cls.inheritted_headers))

    @classmethod
    def should_inherit_header(cls, header_name: str) -> bool:
        return header_name.lower() in cls.inheritted_headers

    @class_property
    def __content_type__(cls) -> ContentType:
        """Content type with `text/plain`.
        """
        return DEFAULT_CONTENT_TYPE_PLAIN

    def get_headers(self) -> t.List[t.Tuple[str, str]]:
        """Publishes additional headers for error response.

        Returns:
            Additional headers.
        """
        return []

    def get_body(self) -> t.Union[bytes, t.Iterable[bytes]]:
        """Publishes response body for error response.

        Returns
            Response body of the error.
        """
        return b""

    def get_all_form(
        self,
    ) -> t.Tuple[
        HTTPStatus,
        t.List[t.Tuple[str, str]],
        t.Union[bytes, t.Iterable[bytes]]
    ]:
        """Get status code, headers and body of repsponse of the error.

        Returns:
            Tuple of status code, headers and body of the error.
        """
        stat = self.http_status
        headers = self.get_headers()
        body = self.get_body()

        # NOTE
        #   Automatically Content-Type header is to be added.
        headers.append(self.__content_type__.to_header())
        return (stat, headers, BufferedConcatIterator(body))

# ----------------------------------------------------------------------------

# Default errors    ----------------------------------------------------------

class DefaultNotFoundErrInfo(ErrInfo):

    http_status = HTTPStatus.NOT_FOUND


class DefaultDataFormatErrInfo(ErrInfo):

    http_status = HTTPStatus.UNSUPPORTED_MEDIA_TYPE


class DefaultHeaderNotFoundErrInfo(ErrInfo):

    http_status = HTTPStatus.BAD_REQUEST


class DefaultQueryParamNotFoundErrInfo(ErrInfo):

    http_status = HTTPStatus.BAD_REQUEST


class DefaultNotApplicableIpErrInfo(ErrInfo):

    http_status = HTTPStatus.FORBIDDEN


class DefualtCORSErrInfo(ErrInfo):

    http_status = HTTPStatus.FORBIDDEN


_WWW_AUTH_HEADER = "WWW-Authentication"


def get_auth_realm(scheme: str, msg: str) -> str:
    return f'{scheme} realm="{msg}"'


DEFAULT_REALM_MESSAGE = "SECRET AREA"


def get_default_auth_realm(scheme: str) -> str:
    return get_auth_realm(scheme, DEFAULT_REALM_MESSAGE)


class DefaultAuthHeaderNotFoundErrInfo(ErrInfo):

    http_status = HTTPStatus.UNAUTHORIZED

    def __init__(self, scheme: str, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._scheme = scheme

    @property
    def scheme(self) -> str:
        return self._scheme

    def get_headers(self) -> t.List[t.Tuple[str, str]]:
        value = get_default_auth_realm(self._scheme)
        return [(_WWW_AUTH_HEADER, value)]


DEFAULT_NOT_FOUND_ERROR = DefaultNotFoundErrInfo()
DEFUALT_INCORRECT_DATA_FORMAT_ERROR = DefaultDataFormatErrInfo()
DEFAULT_HEADER_NOT_FOUND_ERROR = DefaultHeaderNotFoundErrInfo()
DEFAULT_QUERY_PARAM_NOT_FOUND_ERROR = DefaultQueryParamNotFoundErrInfo()
DEFAULT_NOT_APPLICABLE_IP_ERROR = DefaultNotApplicableIpErrInfo()
DEFAULT_BASIC_AUTH_HEADER_NOT_FOUND_ERROR = DefaultAuthHeaderNotFoundErrInfo(AuthSchemes.basic)
DEFAULT_BEARER_AUTH_HEADER_NOT_FOUND_ERROR = DefaultAuthHeaderNotFoundErrInfo(AuthSchemes.bearer)
DEFAULT_CORS_ERROR = DefualtCORSErrInfo()

# ----------------------------------------------------------------------------

class ApiErrInfo(ErrInfo):
    """ErrInfo to handle API error.

    This ErrInfo has implemented method of 'get_body'. This class
    emits Json data including error information defined by developer
    when the error is sent.

    Attributes:
        http_status (HTTPStatus): HTTP status of the error.
        code (str): Error code for your API.
        dev_message (Optional[str]): Message to explain developers the error.
        user_message (Optional[str]): Message to explain end users the error.
        info (Optional[str]): Information about the error.
        encoding (str): Encoding to encode response body.
    """
    code: str
    dev_message: t.Optional[str] = None
    user_message: t.Optional[str] = None
    info: t.Optional[str] = None
    encoding: str = "UTF-8"

    @class_property
    def __content_type__(cls) -> ContentType:
        """Content type with `application/json` and `charset` specified as
        the value of `encoding`.
        """
        return ContentType(MediaTypes.json, charset=cls.encoding)

    def get_body(self) -> t.Optional[bytes]:
        """Publishes response body for error response.

        Returns
        -------
        bytes
            Response body
        """
        body = {
            "code": self.code,
            "developerMessage": self.dev_message,
            "uesrMessage": self.user_message,
            "info": self.info
        }
        return json.dumps(body).encode(encoding=self.encoding)



from __future__ import annotations
import codecs
import typing as t

from .error import ErrInfo
from .http import HTTPStatus
from .io import BufferedConcatIterator


class _ASGIProtocols:
    """Iterable for the ASGI protocols.
    """

    http        = "http"
    websocket   = "websocket"
    lifespan    = "lifespan"

    _events = {
        http,
        websocket,
        lifespan,
    }
    __instance = None

    def __new__(cls) -> _ASGIProtocols:
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)
        return cls.__instance

    def __iter__(self) -> t.Iterator[str]:
        return iter(self._events)

    def __contains__(self, item: str) -> bool:
        return item in self._events


class _ASGIHTTPEvents:
    """Iterable for events of the ASGI HTTP protocol.
    """

    request         = "http.request"
    response_start  = "http.response.start"
    response_body   = "http.response.body"
    disconnect      = "http.disconnect"

    _events = {
        request,
        response_start,
        response_body,
        disconnect,
    }
    __instance = None

    def __new__(cls) -> _ASGIHTTPEvents:
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)
        return cls.__instance

    def __iter__(self) -> t.Iterator[str]:
        return iter(self._events)

    def __contains__(self, item: str) -> bool:
        return item in self._events


class _ASGIWebSocketEvents:
    """Iterable for events of the ASGI WebSocket protocol.
    """

    connect     = "websocket.connect"
    accept      = "websocket.accept"
    receive     = "websocket.receive"
    send        = "websocket.send"
    disconnect  = "websocket.disconnect"
    close       = "websocket.close"

    _events = {
        connect,
        accept,
        receive,
        send,
        disconnect,
        close,
    }
    __instance = None

    def __new__(cls) -> _ASGIWebSocketEvents:
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)
        return cls.__instance

    def __iter__(self) -> t.Iterator[str]:
        return iter(self._events)

    def __contains__(self, item: str) -> bool:
        return item in self._events


class _ASGILifespanEvents:
    """Iterable for events of the ASGI Lifespan protocol.
    """

    startup             = "lifespan.startup"
    startup_complete    = "lifespan.startup.complete"
    startup_failed      = "lifespan.startup.failed"
    shutdown            = "lifespan.shutdown"
    shutdown_complete   = "lifespan.shutdown.complete"
    shutdown_failed     = "lifespan.shutdown.failed"

    _events = {
        startup,
        startup_complete,
        startup_failed,
        shutdown,
        shutdown_complete,
        shutdown_failed,
    }
    __instance = None

    def __new__(cls) -> _ASGILifespanEvents:
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)
        return cls.__instance

    def __iter__(self) -> t.Iterator[str]:
        return iter(self._events)

    def __contains__(self, item: str) -> bool:
        return item in self._events


ASGIProtocols = _ASGIProtocols()
ASGIHTTPEvents = _ASGIHTTPEvents()
ASGIWebSocketEvents = _ASGIWebSocketEvents()
ASGILifespanEvenets = _ASGILifespanEvents()


def _convert_headers_asgistyle(
    headers: t.Iterable[t.Tuple[str, str]]
) -> t.List[t.Tuple[bytes, bytes]]:
    return [tuple(map(codecs.encode, h)) for h in headers]


ASGIRecv_t = t.Callable[
    [],
    t.Awaitable[t.Dict[str, t.Any]]
]
ASGISend_t = t.Callable[
    [t.Dict[str, t.Any]],
    t.Awaitable[None]
]

HTTPSendStart_t = t.Callable[
    [HTTPStatus, t.Iterable[t.Tuple[str, str]]],
    t.Awaitable[None]
]
HTTPSendBody_t = t.Callable[
    [t.Iterable[bytes]],
    t.Awaitable[None]
]
HTTPSendErrInfo_t = t.Callable[
    [ErrInfo],
    t.Awaitable[None]
]


def format_http_sendstart_msg(
    status: HTTPStatus,
    headers: t.Iterable[str, str],
) -> t.Dict[str, t.Any]:
    return {
        "type": ASGIHTTPEvents.response_start,
        "status": status.asgi,
        "headers": _convert_headers_asgistyle(headers),
    }


def format_http_sendbody_msg(
    chunk: bytes = b"",
    more: bool = False,
) -> t.Dict[str, t.Any]:
    return {
        "type": ASGIHTTPEvents.response_body,
        "body": chunk,
        "more_body": more,
    }


def get_http_sendstart(send: ASGISend_t) -> HTTPSendStart_t:

    async def sendstart(
        status: HTTPStatus,
        headers: t.Iterable[t.Tuple[str, str]],
    ) -> None:
        msg = format_http_sendstart_msg(status, headers)
        await send(msg)

    return sendstart


def get_http_sendbody(send: ASGISend_t) -> HTTPSendBody_t:

    async def sendbody(body: t.Union[t.Iterable[bytes]]) -> None:
        for chunk in body:
            msg = format_http_sendbody_msg(chunk, more=True)
            await send(msg)
        msg = format_http_sendbody_msg()
        await send(msg)

    return sendbody


def get_http_send_errinfo(send: ASGISend_t) -> HTTPSendErrInfo_t:
    sendstart = get_http_sendstart(send)
    sendbody = get_http_sendbody(send)

    async def send_errinfo(
        errinfo: ErrInfo,
        res_headers: t.Iterable[t.Tuple[str, str]],
    ) -> None:
        status, headers, body = errinfo.get_all_form()

        # Judge whether the response headers should be inheritted.
        for header_name, header_value in res_headers:
            if errinfo.should_inherit_header(header_name):
                headers.append((header_name, header_value))

        await sendstart(status, headers)
        await sendbody(BufferedConcatIterator(body))

    return send_errinfo


class WebSocketError(Exception):
    pass


class WebSocketDisconnectedError(Exception):
    pass


WebSocketAccept_t = t.Callable[
    [t.Iterable[t.Tuple[str, str]], t.Optional[str]],
    t.Awaitable[None]
]
WebSocketRecvMsg_t = t.Callable[
    [],
    t.Awaitable[t.Tuple[t.Optional[str], t.Optional[bytes]]]
]
WebSocketSendMsg_t = t.Callable[
    [t.Optional[str], t.Optional[bytes]],
    t.Awaitable[None]
]
WebSocketClose_t = t.Callable[
    [int],
    t.Awaitable[None]
]


def format_websock_accept_msg(
    headers: t.Iterable[str, str],
    subprotocol: t.Optional[str] = None,
) -> t.Dict[str, t.Any]:
    return {
        "type": ASGIWebSocketEvents.accept,
        "subprotocol": subprotocol,
        "headers": _convert_headers_asgistyle(headers),
    }


def format_websock_send_msg(
    text: t.Optional[str] = None,
    bin: t.Optional[bytes] = None,
) -> t.Dict[str, t.Any]:
    if text is None and bin is None:
        raise ValueError("One of 'text' or 'bin' must be non-None.")

    return {
        "type": ASGIWebSocketEvents.send,
        "bytes": bin,
        "text": text,
    }


def format_websock_close_msg(code: int = 1000) -> t.Dict[str, t.Any]:
    return {
        "type": ASGIWebSocketEvents.close,
        "code": code,
    }


def get_websock_accept(send: ASGISend_t) -> WebSocketAccept_t:

    async def websock_accept(
        headers: t.Iterable[str, str],
        subprotocol: t.Optional[str] = None,
    ) -> None:
        msg = format_websock_accept_msg(headers, subprotocol)
        await send(msg)

    return websock_accept


def get_websock_recvmsg(recv: ASGIRecv_t) -> WebSocketRecvMsg_t:

    async def websock_recvmsg() -> t.Tuple[str, bytes]:
        msg = await recv()
        if msg.get("type") == ASGIWebSocketEvents.disconnect:
            raise WebSocketDisconnectedError()
        return (msg.get("text"), msg.get("bytes"))

    return websock_recvmsg


def get_websock_sendmsg(send: ASGISend_t) -> WebSocketSendMsg_t:

    async def websock_sendmsg(
        text: t.Optional[str] = None,
        bin: t.Optional[str] = None,
    ) -> None:
        msg = format_websock_send_msg(text, bin)
        await send(msg)

    return websock_sendmsg


def get_websock_close(send: ASGISend_t) -> WebSocketClose_t:

    async def websock_close(code: int = 1000) -> None:
        msg = format_websock_close_msg(code)
        await send(msg)

    return websock_close


LifespanHandler_t = t.Callable[
    [t.Dict[str, t.Any], ASGIRecv_t, ASGISend_t],
    t.Awaitable[None]
]


async def default_lifespan_handler(
    scope: t.Dict[str, t.Any],
    recv: t.Callable[[], t.Awaitable[t.Dict[str, t.Any]]],
    send: t.Callable[[t.Dict[str, t.Any]], t.Awaitable[None]],
) -> None:
    while True:
        msg = await recv()
        msg_typ = msg.get("type")
        if msg_typ == ASGILifespanEvenets.startup:
            await send({"type": ASGILifespanEvenets.startup_complete})
        elif msg_typ == ASGILifespanEvenets.shutdown:
            await send({"type": ASGILifespanEvenets.shutdown_complete})
            return



from __future__ import annotations
from abc import ABCMeta, abstractmethod
import dataclasses
import enum
import re
import typing as t

from .util.deco import class_property
from .util.header import make_header


__all__ = []


class _HTTPMethods:
    """Iterator for HTTP methods.
    """

    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    HEAD = "HEAD"
    OPTIONS = "OPTIONS"
    PATCH = "PATCH"
    TRACE = "TRACE"
    CONNECT = "CONNECT"

    __methods = set((
        GET,
        POST,
        PUT,
        DELETE,
        HEAD,
        OPTIONS,
        PATCH,
        TRACE,
    ))

    __instance = None

    def __new__(cls) -> _HTTPMethods:
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)
        return cls.__instance

    def __iter__(self) -> t.Iterable[str]:
        return iter(self.__methods)

    def __contains__(self, item: str) -> bool:
        return item in self.__methods


class _MediaTypes:
    """Iterator for media types of body on communication.
    """

    plain = "text/plain"
    html = "text/html"
    xml = "application/xml"
    css = "text/css"
    javascript = "application/javascript"
    json = "application/json"
    x_www_form_urlencoded = "application/x-www-form-urlencoded"
    rss = "application/rss+xml"
    atom = "application/atom+xml"
    binary = "application/octet-stream"
    zip = "application/zip"
    jpeg = "image/jpeg"
    png = "image/png"
    svg = "image/svg+xml"
    multi_form = "multipart/form-data"
    mp4 = "video/mp4"
    excel = "application/vnd.ms-excel"

    __types = set((
        plain,
        html,
        xml,
        css,
        javascript,
        json,
        rss,
        atom,
        binary,
        zip,
        jpeg,
        png,
        svg,
        multi_form,
        mp4,
        excel,
    ))

    __instance = None

    def __new__(cls) -> _MediaTypes:
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)
        return cls.__instance

    def __iter__(self) -> t.Iterable[str]:
        return iter(self.__types)

    def __contains__(self, item: str) -> bool:
        return item in self.__types


class _AuthSchemes:
    """Iterator for authentication schemes of the Authorization header.
    """

    basic = "Basic"
    bearer = "Bearer"

    __schemes = set((
        basic,
        bearer
    ))

    __instance = None

    def __new__(cls) -> _AuthSchemes:
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)
        return cls.__instance

    def __iter__(self) -> t.Iterable[str]:
        return iter(self.__schemes)

    def __contains__(self, item: str) -> bool:
        return item in self.__schemes


HTTPMethods = _HTTPMethods()
MediaTypes = _MediaTypes()
AuthSchemes = _AuthSchemes()


@dataclasses.dataclass
class ContentType:
    """`dataclass` for describing value of `Content-Type` header.

    Its object is often used when handling the value of `Content-Type` in
    Bamboo.
    """

    media_type: str
    charset: t.Optional[str] = None
    boundary: t.Optional[str] = None

    def to_header(self) -> t.Tuple[str, str]:
        """Format contents as a Content-Type header.

        Returns:
            Header name and value of the header.
        """
        params = {}
        if self.charset:
            params["charset"] = self.charset
        if self.boundary:
            params["boundary"] = self.boundary
        return make_header("Content-Type", self.media_type, **params)

    @classmethod
    def parse(cls, raw: str) -> ContentType:
        """Parse and make new instance of this class.

        Parameters
        ----------
        raw : str
            Value of `Content-Type` header

        Returns
        -------
        ContentType
            New instance of this class based on the `raw` data
        """
        raw = re.split(";\s*", raw)
        result = cls(raw[0])

        for directive, val in [item.split("=") for item in raw[1:]]:
            if directive == "charset":
                result.charset = val.lower()
            elif directive == "boundary":
                result.boundary = val

        return result


class ContentTypeHolder(metaclass=ABCMeta):
    """Abstract class with properties about `Content-Type`.
    """

    @class_property
    @abstractmethod
    def __content_type__(cls) -> ContentType:
        pass

    @classmethod
    def verify_content_type(cls, content_type: ContentType) -> bool:
        media_type_expected = cls.__content_type__.media_type.lower()
        media_type_verified = content_type.media_type.lower()
        if media_type_expected != media_type_verified:
            return False

        if cls.__content_type__.charset:
            charset_expected = cls.__content_type__.charset.lower()
            charset_verified = content_type.charset.lower()
            if charset_expected != charset_verified:
                return False
        return True


# NOTE
#   Value of the Content-Type header by default. This variables can be
#   used if Content-Type of response is not specified.
DEFAULT_CONTENT_TYPE_PLAIN = ContentType(MediaTypes.plain)
DEFAULT_CONTENT_TYPE_JSON = ContentType(MediaTypes.json, "UTF-8")


class HTTPStatus(int, enum.Enum):
    """Enum which describes HTTP status.

    Each class variables of the class have four main attributes:

    - value
    - wsgi
    - asgi
    - description

    `value` means the enum value and describes HTTP status as int.
    `wsgi` is str and used only for WSGI servers.
    Similarly, `asgi` is int and only for ASGI servers.
    'description' describes short message of corresponding HTTP satus.
    """

    def __new__(cls, value: int, wsgi: str, asgi: int, description: str = ""):
        obj = int.__new__(cls, value)
        obj._value_ = value
        obj.wsgi = wsgi
        obj.asgi = asgi
        obj.description = description
        return obj

    @classmethod
    def get_status(cls, code: int) -> t.Optional[HTTPStatus]:
        for stat in cls:
            if stat.value == code:
                return stat
        return None

    CONTINUE = (
        100,
        "100 Continue",
        100,
        "Request received, please continue"
    )
    SWITCHING_PROTOCOLS = (
        101,
        "101 Switching Protocols",
        101,
        "Switching to new protocol; obey Upgrade header"
    )
    PROCESSING = (
        102,
        "102 Processing",
        102,
        ""
    )

    # success
    OK  = (
        200,
        "200 OK",
        200,
        "Request fulfilled, document follows"
    )
    CREATED = (
        201,
        "201 Created",
        201,
        "Document created, URL follows"
    )
    ACCEPTED = (
        202,
        "202 Accepted",
        202,
        "Request accepted, processing continues off-line"
    )
    NON_AUTHORITATIVE_INFORMATION = (
        203,
        "203 Non-Authoritative Information",
        203,
        "Request fulfilled from cache"
    )
    NO_CONTENT = (
        204,
        "204 No Content",
        204,
        "Request fulfilled, nothing follows"
    )
    RESET_CONTENT = (
        205,
        "205 Reset Content",
        205,
        "Clear input form for further input"
    )
    PARTIAL_CONTENT = (
        206,
        "206 Partial Content",
        206,
        "Partial content follows"
    )
    MULTI_STATUS = (
        207,
        "207 Multi-Status",
        207,
        ""
    )
    ALREADY_REPORTED = (
        208,
        "208 Already Reported",
        208,
        ""
    )
    IM_USED = (
        226,
        "226 IM Used",
        226,
        ""
    )

    # redirection
    MULTIPLE_CHOICES = (
        300,
        "300 Multiple Choices",
        300,
        "Object has several resources -- see URI list"
    )
    MOVED_PERMANENTLY = (
        301,
        "301 Moved Permanently",
        301,
        "Object moved permanently -- see URI list"
    )
    FOUND = (
        302,
        "302 Found",
        302,
        "Object moved temporarily -- see URI list"
    )
    SEE_OTHER = (
        303,
        "303 See Other",
        303,
        "Object moved -- see Method and URL list"
    )
    NOT_MODIFIED = (
        304,
        "304 Not Modified",
        304,
        "Document has not changed since given time"
    )
    USE_PROXY = (
        305,
        "305 Use Proxy",
        305,
        "You must use proxy specified in Location to access this resource"
    )
    TEMPORARY_REDIRECT = (
        307,
        "307 Temporary Redirect",
        307,
        "Object moved temporarily -- see URI list"
    )
    PERMANENT_REDIRECT = (
        308,
        "308 Permanent Redirect",
        308,
        "Object moved permanently -- see URI list"
    )

    # client error
    BAD_REQUEST = (
        400,
        "400 Bad Request",
        400,
        "Bad request syntax or unsupported method"
    )
    UNAUTHORIZED = (
        401,
        "401 Unauthorized",
        401,
        "No permission -- see authorization schemes"
    )
    PAYMENT_REQUIRED = (
        402,
        "402 Payment Required",
        402,
        "No payment -- see charging schemes"
    )
    FORBIDDEN = (
        403,
        "403 Forbidden",
        403,
        "Request forbidden -- authorization will not help"
    )
    NOT_FOUND = (
        404,
        "404 Not Found",
        404,
        "Nothing matches the given URI"
    )
    METHOD_NOT_ALLOWED = (
        405,
        "405 Method Not Allowed",
        405,
        "Specified method is invalid for this resource"
    )
    NOT_ACCEPTABLE = (
        406,
        "406 Not Acceptable",
        406,
        "URI not available in preferred format"
    )
    PROXY_AUTHENTICATION_REQUIRED = (
        407,
        "407 Proxy Authentication Required",
        407,
        "You must authenticate with this proxy before proceeding"
    )
    REQUEST_TIMEOUT = (
        408,
        "408 Request Timeout",
        408,
        "Request timed out; try again later"
    )
    CONFLICT = (
        409,
        "409 Conflict",
        409,
        "Request conflict"
    )
    GONE = (
        410,
        "410 Gone",
        410,
        "URI no longer exists and has been permanently removed"
    )
    LENGTH_REQUIRED = (
        411,
        "411 Length Required",
        411,
        "Client must specify Content-Length"
    )
    PRECONDITION_FAILED = (
        412,
        "412 Precondition Failed",
        412,
        "Precondition in headers is false"
    )
    REQUEST_ENTITY_TOO_LARGE = (
        413,
        "413 Request Entity Too Large",
        413,
        "Entity is too large"
    )
    REQUEST_URI_TOO_LONG = (
        414,
        "414 Request-URI Too Long",
        414,
        "URI is too long"
    )
    UNSUPPORTED_MEDIA_TYPE = (
        415,
        "415 Unsupported Media Type",
        415,
        "Entity body in unsupported format"
    )
    REQUESTED_RANGE_NOT_SATISFIABLE = (
        416,
        "416 Requested Range Not Satisfiable",
        416,
        "Cannot satisfy request range"
    )
    EXPECTATION_FAILED = (
        417,
        "417 Expectation Failed",
        417,
        "Expect condition could not be satisfied"
    )
    MISDIRECTED_REQUEST = (
        421,
        "421 Misdirected Request",
        421,
        "Server is not able to produce a response"
    )
    UNPROCESSABLE_ENTITY = (
        422,
        "422 Unprocessable Entity",
        422,
        ""
    )
    LOCKED = (
        423,
        "423 Locked",
        423,
        ""
    )
    FAILED_DEPENDENCY = (
        424,
        "424 Failed Dependency",
        424,
        ""
    )
    UPGRADE_REQUIRED = (
        426,
        "426 Upgrade Required",
        426,
        ""
    )
    PRECONDITION_REQUIRED = (
        428,
        "428 Precondition Required",
        428,
        "The origin server requires the request to be conditional"
    )
    TOO_MANY_REQUESTS = (
        429,
        "429 Too Many Requests",
        429,
        "The user has sent too many requests in a given amount of time"
        "('rate limiting')"
    )
    REQUEST_HEADER_FIELDS_TOO_LARGE = (
        431,
        "431 Request Header Fields Too Large",
        431,
        "The server is unwilling to process the request because its "
        "header fields are too large"
    )

    # server errors
    INTERNAL_SERVER_ERROR = (
        500,
        "500 Internal Server Error",
        500,
        "Server got itself in trouble"
    )
    NOT_IMPLEMENTED = (
        501,
        "501 Not Implemented",
        501,
        "Server does not support this operation"
    )
    BAD_GATEWAY = (
        502,
        "502 Bad Gateway",
        502,
        "Invalid responses from another server/proxy"
    )
    SERVICE_UNAVAILABLE = (
        503,
        "503 Service Unavailable",
        503,
        "The server cannot process the request due to a high load"
    )
    GATEWAY_TIMEOUT = (
        504,
        "504 Gateway Timeout",
        504,
        "The gateway server did not receive a timely response"
    )
    HTTP_VERSION_NOT_SUPPORTED = (
        505,
        "505 HTTP Version Not Supported",
        505,
        "Cannot fulfill request"
    )
    VARIANT_ALSO_NEGOTIATES = (
        506,
        "506 Variant Also Negotiates",
        506,
        ""
    )
    INSUFFICIENT_STORAGE = (
        507,
        "507 Insufficient Storage",
        507,
        ""
    )
    LOOP_DETECTED = (
        508,
        "508 Loop Detected",
        508,
        ""
    )
    NOT_EXTENDED = (
        510,
        "510 Not Extended",
        510,
        ""
    )
    NETWORK_AUTHENTICATION_REQUIRED = (
        511,
        "511 Network Authentication Required",
        511,
        "The client needs to authenticate to gain network access"
    )



from __future__ import annotations
import os
import urllib.parse
import typing as t

from ..api.base import ApiData


__all__ = [
    "Response",
    "Schemes",
]


ResponseData_t = t.TypeVar("ResponseData_t", bound=ApiData)


class _Schemes:

    HTTP = "http"
    HTTPS = "https"

    _schemes = set((HTTP, HTTPS))
    __instance = None

    def __new__(cls) -> _Schemes:
        if cls.__instance is None:
            cls.__instance = super().__new__(cls)
        return cls.__instance

    def __iter__(self):
        return iter(self._schemes)

    def __contains__(self, item: str):
        return item in self._schemes


Schemes = _Schemes()


def _get_http_proxy_env() -> t.Optional[str]:
    env = os.environ.get("http_proxy")
    if env is None:
        env = os.environ.get("HTTP_PROXY")
    return env


def _get_https_proxy_env() -> t.Optional[str]:
    env = os.environ.get("https_proxy")
    if env is None:
        env = os.environ.get("HTTPS_PROXY")
    return env


def _parse_proxy_netloc(uri: str) -> t.Tuple[str, int]:
    result = urllib.parse.urlparse(uri)
    try:
        domain, port = result.netloc.split(":", 2)
    except ValueError as e:
        domain = result.netloc
        port = 80
    return (domain, int(port))


from bamboo.request.response import Response



from abc import ABCMeta, abstractmethod
import dataclasses
import functools
import inspect
import re
import typing as t

from . import (
    Callback_ASGI_t,
    Callback_WSGI_t,
    Callback_t,
    CallbackDecorator_t,
    DuplicatedInfoError,
    _get_bamboo_attr,
)
from ..api.base import ApiData, ApiValidationFailedError
from ..endpoint import (
    ASGIEndpointBase,
    ASGIHTTPEndpoint,
    HTTPMixIn,
    WSGIEndpoint,
    set_response_method,
)
from ..error import (
    DEFAULT_BASIC_AUTH_HEADER_NOT_FOUND_ERROR,
    DEFAULT_BEARER_AUTH_HEADER_NOT_FOUND_ERROR,
    DEFAULT_HEADER_NOT_FOUND_ERROR,
    DEFAULT_CORS_ERROR,
    DEFUALT_INCORRECT_DATA_FORMAT_ERROR,
    DEFAULT_NOT_APPLICABLE_IP_ERROR,
    ErrInfo,
)
from ..http import AuthSchemes, HTTPStatus
from ..util.convert import decode2binary
from ..util.ip import is_valid_ipv4


__all__ = [
    "SetCookieValue_t",
    "add_preflight",
    "allow_simple_access_control",
    "basic_auth",
    "bearer_auth",
    "data_format",
    "get_asgi_preflight",
    "get_wsgi_preflight",
    "has_header_of",
    "has_query_of",
    "may_occur",
    "restricts_client",
    "set_cache_control",
    "set_cookie",
]


class CallbackConfigBase(metaclass=ABCMeta):

    ATTR: str

    @abstractmethod
    def get(self) -> t.Any:
        pass

    @abstractmethod
    def set(self, *args, **kwargs) -> Callback_t:
        pass


class HTTPEndpointConfigBase(metaclass=ABCMeta):

    ATTR: str


    @abstractmethod
    def get(self) -> t.Any:
        pass

    @abstractmethod
    def set(self, *args, **kwargs) -> HTTPMixIn:
        pass


class HTTPErrorConfig(CallbackConfigBase):

    ATTR = _get_bamboo_attr("errors")

    def __init__(self, callback) -> None:
        if not hasattr(callback, self.ATTR):
            setattr(callback, self.ATTR, set())

        self._callback = callback
        self._registered: t.Set[ErrInfo] = getattr(callback, self.ATTR)

    def get(self) -> t.Tuple[t.Type[ErrInfo], ...]:
        return tuple(self._registered)

    def set(self, *errors: t.Type[ErrInfo]) -> Callback_t:
        for err in errors:
            self._registered.add(err)
        return self._callback


def may_occur(*errors: t.Type[ErrInfo]) -> CallbackDecorator_t:
    """Register error classes to callback on `Endpoint`.

    Args:
        *errors: Error classes which may occuur.

    Returns:
        CallbackDecorator_t: Decorator to register
            error classes to callback.

    Examples:
        ```python
        class MockErrInfo(ErrInfo):
            http_status = HTTPStatus.INTERNAL_SERVER_ERROR

            def get_body(self) -> bytes:
                return b"Intrernal server error occured"

        class MockEndpoint(WSGIEndpoint):

            @may_occur(MockErrInfo)
            def do_GET(self) -> None:
                # Do something...

                # It is possible to send error response.
                if is_some_flag():
                    self.send_err(MockErrInfo())

                self.send_only_status()
        ```
    """

    def wrapper(callback: Callback_t) -> Callback_t:
        config = HTTPErrorConfig(callback)
        return config.set(*errors)

    return wrapper


@dataclasses.dataclass(eq=True, frozen=True)
class DataFormatInfo:
    """`dataclass` with information of data format at callbacks on `Endpoint`.

    Attributes:
        input (Optional[Type[ApiData]]): Input data format.
        output (Optional[Type[ApiData]]): Output data format.
        is_validate (bool): If input data is to be validate.
        err_validate (ErrInfo): Error information sent
            when validation failes.
    """
    input: t.Optional[t.Type[ApiData]] = None
    output: t.Optional[t.Type[ApiData]] = None
    is_validate: bool = True
    err_validate: ErrInfo = DEFUALT_INCORRECT_DATA_FORMAT_ERROR
    err_noheader: ErrInfo = DEFAULT_HEADER_NOT_FOUND_ERROR


class DataFormatConfig(CallbackConfigBase):

    ATTR = _get_bamboo_attr("data_format")

    def __init__(self, callback: Callback_t) -> None:
        super().__init__()

        self._callback = callback

    def get(self) -> t.Optional[DataFormatInfo]:
        return getattr(self._callback, self.ATTR, None)

    def set(self, dataformat: DataFormatInfo) -> Callback_t:
        setattr(self._callback, self.ATTR, dataformat)
        if not dataformat.is_validate:
            return self._callback

        if inspect.iscoroutinefunction(self._callback):
            if dataformat.input:
                func = self.decorate_asgi
            else:
                func = self.decorate_asgi_no_input
        else:
            if dataformat.input:
                func = self.decorate_wsgi
            else:
                func = self.decorate_wsgi_no_input

        return func(self._callback, dataformat)

    @staticmethod
    def decorate_wsgi(
        callback: Callback_WSGI_t,
        dataformat: DataFormatInfo
    ) -> Callback_WSGI_t:

        @functools.wraps(callback)
        @may_occur(dataformat.err_validate.__class__)
        @has_header_of("Content-Type", dataformat.err_noheader, add_arg=False)
        def _callback(self: WSGIEndpoint, *args) -> None:
            body = self.body
            try:
                data = dataformat.input.__validate__(body, self.content_type)
            except ApiValidationFailedError:
                raise dataformat.err_validate
            callback(self, data, *args)

        return _callback

    @staticmethod
    def decorate_wsgi_no_input(
        callback: Callback_WSGI_t,
        dataformat: DataFormatInfo
    ) -> Callback_WSGI_t:

        @functools.wraps(callback)
        def _callback(self: WSGIEndpoint, *args) -> None:
            setattr(self, "body", b"")
            callback(self, *args)

        return _callback

    @staticmethod
    def decorate_asgi(
        callback: Callback_ASGI_t,
        dataformat: DataFormatInfo
    ) -> Callback_ASGI_t:

        @functools.wraps(callback)
        @may_occur(dataformat.err_validate.__class__)
        @has_header_of("Content-Type", dataformat.err_noheader, add_arg=False)
        async def _callback(self: ASGIHTTPEndpoint, *args) -> None:
            body = await self.body
            try:
                data = dataformat.input.__validate__(body, self.content_type)
            except ApiValidationFailedError:
                raise dataformat.err_validate
            await callback(self, data, *args)

        return _callback

    @staticmethod
    def decorate_asgi_no_input(
        callback: Callback_ASGI_t,
        dataformat: DataFormatInfo
    ) -> Callback_ASGI_t:

        @functools.wraps(callback)
        async def _callback(self: ASGIHTTPEndpoint, *args) -> None:
            # NOTE
            #   Reference awaitable_cached_property
            body_property = await self.__class__.body
            body_property._obj2val[self] = b""
            await callback(self, *args)

        return _callback


def data_format(
    input: t.Optional[t.Type[ApiData]] = None,
    output: t.Optional[t.Type[ApiData]] = None,
    is_validate: bool = True,
    err_validate: ErrInfo = DEFUALT_INCORRECT_DATA_FORMAT_ERROR,
    err_noheader: ErrInfo = DEFAULT_HEADER_NOT_FOUND_ERROR,
) -> CallbackDecorator_t:
    """Set data format of input/output data as API to callback on
    `Endpoint`.

    This decorator can be used to add attributes of data format information
    to a response method, and execute validation if input raw data has
    expected format defined on `input` argument.

    To represent no data inputs/outputs, specify `input`/`output` arguments
    as `None`. If `input` is `None`, then any data received from client will
    not be read. If `is_validate` is `False`, then validation will not be
    executed.

    Args:
        input: Input data format.
        output: Output data format.
        is_validate: If input data is to be validated.
        err_validate: Error information sent when validation failes.

    Returns:
        Decorator to add attributes of data format information to callback.

    Examples:
        ```python
        class UserData(JsonApiData):
            name: str
            email: str
            age: int

        class MockEndpoint(WSGIEndpoint):

            @data_format(input=UserData, output=None)
            def do_GET(self, rec_body: UserData) -> None:
                user_name = rec_body.name
                # Do something...
        ```
    """
    dataformat = DataFormatInfo(
        input,
        output,
        is_validate,
        err_validate,
        err_noheader,
    )

    def wrapper(callback: Callback_t) -> Callback_t:
        config = DataFormatConfig(callback)
        return config.set(dataformat)

    return wrapper


@dataclasses.dataclass(eq=True, frozen=True)
class RequiredHeaderInfo:
    """`dataclass` with information of header which should be included in
    response headers.

    Attributes:
        header: Name of header.
        err: Error information sent when the header is not included.
        add_arg: Whether the header is given as a callback's argument.
    """
    header: str
    err: t.Optional[ErrInfo]
    add_arg: bool


class RequiredHeaderConfig(CallbackConfigBase):

    ATTR = _get_bamboo_attr("required_headers")

    def __init__(self, callback: Callback_t) -> None:
        super().__init__()

        if not hasattr(callback, self.ATTR):
            setattr(callback, self.ATTR, set())

        self._callback = callback
        self._registered: t.Set[RequiredHeaderInfo] = getattr(callback, self.ATTR)

    def get(self) -> t.Tuple[RequiredHeaderInfo]:
        return tuple(self._registered)

    def set(self, info: RequiredHeaderInfo) -> Callback_t:
        self._registered.add(info)

        if inspect.iscoroutinefunction(self._callback):
            func = self.decorate_asgi
        else:
            func = self.decorate_wsgi

        return func(self._callback, info)

    @staticmethod
    def decorate_wsgi(
        callback: Callback_WSGI_t,
        info: RequiredHeaderInfo,
    ) -> Callback_WSGI_t:

        @functools.wraps(callback)
        def _callback(self: WSGIEndpoint, *args) -> None:
            val = self.get_header(info.header)
            if val is None and info.err:
                raise info.err

            if info.add_arg:
                callback(self, val, *args)
            else:
                callback(self, *args)

        if info.err:
            _callback = may_occur(info.err.__class__)(_callback)
        return _callback

    @staticmethod
    def decorate_asgi(
        callback: Callback_ASGI_t,
        info: RequiredHeaderInfo,
    ) -> Callback_ASGI_t:

        @functools.wraps(callback)
        async def _callback(self: ASGIHTTPEndpoint, *args) -> None:
            val = self.get_header(info.header)
            if val is None and info.err:
                raise info.err

            if info.add_arg:
                await callback(self, val, *args)
            else:
                await callback(self, *args)

        if info.err:
            _callback = may_occur(info.err.__class__)(_callback)
        return _callback


def has_header_of(
    header: str,
    err: t.Optional[ErrInfo] = None,
    add_arg: bool = True,
) -> CallbackDecorator_t:
    """Set callback up to receive given header from clients.

    If request headers don't include specified `header`, then response
    headers and body will be made based on `err` and sent.

    Args:
        header: Name of header.
        err: Error information sent when specified `header` is not found.
        add_arg: Whether the header is given as a callback's argument.

    Returns:
        Decorator to make callback to be set up to receive the header.

    Examples:
        ```python
        class BasicAuthHeaderNotFoundErrInfo(ErrInfo):
            http_status = HTTPStatus.UNAUTHORIZED

            def get_headers(self) -> List[Tuple[str, str]]:
                return [("WWW-Authenticate", 'Basic realm="SECRET AREA"')]

        class MockEndpoint(WSGIEndpoint):

            @has_header_of("Authorization", BasicAuthHeaderNotFoundErrInfo())
            def do_GET(self) -> None:
                # It is guaranteed that request headers include the
                # `Authorization` header at this point.
                header_auth = self.get_header("Authorization")

                # Do something...
        ```
    """
    info = RequiredHeaderInfo(header, err, add_arg)

    def wrapper(callback: Callback_t) -> Callback_t:
        config = RequiredHeaderConfig(callback)
        return config.set(info)

    return wrapper


@dataclasses.dataclass(eq=True, frozen=True)
class ClientInfo:

    ip: str
    port: t.Optional[int] = None


_RestrictedClient_t = t.Dict[str, t.Set[t.Optional[int]]]


class RestrictedClientsConfig(CallbackConfigBase):

    ATTR = _get_bamboo_attr("restricted_clients")

    def __init__(self, callback: Callback_t) -> None:
        super().__init__()

        if not hasattr(callback, self.ATTR):
            setattr(callback, self.ATTR, {})

        self._callback = callback
        self._registered: _RestrictedClient_t = getattr(callback, self.ATTR)

    def get(self) -> _RestrictedClient_t:
        return self._registered.copy()

    def set(
        self,
        *clients: ClientInfo,
        err: ErrInfo = DEFAULT_NOT_APPLICABLE_IP_ERROR
    ) -> Callback_t:
        for client in clients:
            if not is_valid_ipv4(client.ip):
                raise ValueError(f"{client.ip} is an invalid IP address.")
            if client.ip == "localhost":
                client = ClientInfo("127.0.0.1", client.port)

            ports = self._registered.get(client.ip)
            if ports is None:
                self._registered[client.ip] = set()
            self._registered[client.ip].add(client.port)

        if inspect.iscoroutinefunction(self._callback):
            func = self.decorate_asgi
        else:
            func = self.decorate_wsgi
        return func(self._callback, err)

    @classmethod
    def decorate_wsgi(
        cls,
        callback: Callback_WSGI_t,
        err: ErrInfo
    ) -> Callback_WSGI_t:
        acceptables = getattr(callback, cls.ATTR, {})

        @functools.wraps(callback)
        @may_occur(err.__class__)
        def _callback(self: WSGIEndpoint, *args) -> None:
            client = ClientInfo(*self.get_client_addr())
            ports = acceptables.get(client.ip)
            if ports is None:
                raise err
            if not(None in ports or client.port in ports):
                raise err
            callback(self, *args)

        return _callback

    @classmethod
    def decorate_asgi(
        cls,
        callback: Callback_ASGI_t,
        err: ErrInfo
    ) -> Callback_ASGI_t:
        acceptables = getattr(callback, cls.ATTR, set())

        @functools.wraps(callback)
        @may_occur(err.__class__)
        async def _callback(self: ASGIHTTPEndpoint, *args) -> None:
            client = ClientInfo(*self.get_client_addr())
            ports = acceptables.get(client.ip)
            if ports is None:
                raise err
            if not(None in ports or client.port in ports):
                raise err
            await callback(self, *args)

        return _callback


def restricts_client(
    *clients: ClientInfo,
    err: ErrInfo = DEFAULT_NOT_APPLICABLE_IP_ERROR
) -> CallbackDecorator_t:
    """Restrict IP addresses at callback.

    Args:
        *client_ips: IP addresses to be allowed to request
        err: Error information sent when request from IP not included
            specified IPs comes.

    Returns:
        Decorator to make callback to be set up to restrict IP addresses.

    Raises:
        ValueError: Raised if invalid IP address is detected

    Examples:
        ```python
        class MockEndpoint(WSGIEndpoint):

            # Restrict to allow only localhost to request
            # to this callback
            @restricts_client(ClientInfo("localhost"))
            def do_GET(self) -> None:
                # Only localhost can access to the callback.

                # Do something...
        ```
    """
    def wrapper(callback: Callback_t) -> Callback_t:
        config = RestrictedClientsConfig(callback)
        return config.set(*clients, err=err)

    return wrapper


class MultipleAuthSchemeError(Exception):
    """Raised if several authentication schemes of the 'Authorization'
    header are detected.
    """
    pass


class AuthSchemeConfig(CallbackConfigBase):

    ATTR = _get_bamboo_attr("auth_scheme")
    HEADER_AUTHORIZATION = "Authorization"

    def __init__(self, callback: Callback_t) -> None:
        super().__init__()

        self._callback = callback

    def get(self) -> t.Optional[str]:
        return getattr(self._callback, self.ATTR, None)

    def set(self, scheme: str, err: ErrInfo) -> Callback_t:
        if hasattr(self._callback, self.ATTR):
            _scheme_registered = getattr(self._callback, self.ATTR)
            raise MultipleAuthSchemeError(
                "Authentication scheme has already been specified as "
                f"'{_scheme_registered}'. Do not specify multiple schemes."
            )

        if scheme not in AuthSchemes:
            raise ValueError(f"Specified scheme '{scheme}' is not supported.")

        setattr(self._callback, self.ATTR, scheme)

        if scheme == AuthSchemes.basic:
            if inspect.iscoroutinefunction(self._callback):
                func = self.decorate_asgi_basic
            else:
                func = self.decorate_wsgi_basic
        elif scheme == AuthSchemes.bearer:
            if inspect.iscoroutinefunction(self._callback):
                func = self.decorate_asgi_bearer
            else:
                func = self.decorate_wsgi_bearer

        return func(self._callback, err)

    @staticmethod
    def _validate_auth_header(value: str, scheme: str) -> t.Optional[str]:
        value = value.split(" ")
        if len(value) != 2:
            return None

        _scheme, credentials = value
        if _scheme != scheme:
            return None

        return credentials

    @classmethod
    def decorate_wsgi_basic(
        cls,
        callback: Callback_WSGI_t,
        err: ErrInfo
    ) -> Callback_WSGI_t:

        @functools.wraps(callback)
        @may_occur(err.__class__)
        @has_header_of(cls.HEADER_AUTHORIZATION, err, add_arg=False)
        def _callback(self: WSGIEndpoint, *args) -> None:
            val = self.get_header(cls.HEADER_AUTHORIZATION)
            credentials = cls._validate_auth_header(val, AuthSchemes.basic)
            if credentials is None:
                raise err

            credentials = decode2binary(credentials).decode().split(":")
            if len(credentials) != 2:
                raise err

            user_id, pw = credentials
            callback(self, user_id, pw, *args)

        return _callback

    @classmethod
    def decorate_asgi_basic(
        cls,
        callback: Callback_ASGI_t,
        err: ErrInfo
    ) -> Callback_ASGI_t:

        @functools.wraps(callback)
        @may_occur(err.__class__)
        @has_header_of(cls.HEADER_AUTHORIZATION, err, add_arg=False)
        async def _callback(self: ASGIHTTPEndpoint, *args) -> None:
            val = self.get_header(cls.HEADER_AUTHORIZATION)
            credentials = cls._validate_auth_header(val, AuthSchemes.basic)
            if credentials is None:
                raise err

            credentials = decode2binary(credentials).decode().split(":")
            if len(credentials) != 2:
                raise err

            user_id, pw = credentials
            await callback(self, user_id, pw, *args)

        return _callback

    @classmethod
    def decorate_wsgi_bearer(
        cls,
        callback: Callback_WSGI_t,
        err: ErrInfo,
    ) -> Callback_WSGI_t:

        @functools.wraps(callback)
        @may_occur(err.__class__)
        @has_header_of(cls.HEADER_AUTHORIZATION, err, add_arg=False)
        def _callback(self: WSGIEndpoint, *args) -> None:
            val = self.get_header(cls.HEADER_AUTHORIZATION)
            token = cls._validate_auth_header(val, AuthSchemes.bearer)
            if token is None:
                raise err
            callback(self, token, *args)

        return _callback

    @classmethod
    def decorate_asgi_bearer(
        cls,
        callback: Callback_ASGI_t,
        err: ErrInfo,
    ) -> Callback_ASGI_t:

        @functools.wraps(callback)
        @may_occur(err.__class__)
        @has_header_of(cls.HEADER_AUTHORIZATION, err, add_arg=False)
        async def _callback(self: ASGIHTTPEndpoint, *args) -> None:
            val = self.get_header(cls.HEADER_AUTHORIZATION)
            token = cls._validate_auth_header(val, AuthSchemes.bearer)
            if token is None:
                raise err
            await callback(self, token, *args)

        return _callback


def basic_auth(
    err: ErrInfo = DEFAULT_BASIC_AUTH_HEADER_NOT_FOUND_ERROR,
) -> CallbackDecorator_t:
    """Set callback up to require `Authorization` header in Basic
    authentication.

    Args:
        err: Error sent when `Authorization` header is not found, received
            scheme doesn't match, or extracting user ID and password from
            credentials failes.

    Returns:
        Decorator to make callback to be set  up to require
        the `Authorization` header.

    Examples:
        ```python
        class MockEndpoint(WSGIEndpoint):

            @basic_auth()
            def do_GET(self, user_id: str, password: str) -> None:
                # It is guaranteed that request headers include the
                # `Authorization` header at this point, and user_id and
                # password are the ones extracted from the header.

                # Authenticate with any function working on your system.
                authenticate(user_id, password)

                # Do something...
        ```
    """
    def wrapper(callback: Callback_t) -> Callback_t:
        config = AuthSchemeConfig(callback)
        return config.set(AuthSchemes.basic, err)

    return wrapper


def bearer_auth(
    err: ErrInfo = DEFAULT_BEARER_AUTH_HEADER_NOT_FOUND_ERROR,
) -> CallbackDecorator_t:
    """Set callback up to require `Authorization` header in token
    authentication for OAuth 2.0.

    Args:
        err : Error sent when `Authorization` header is not found, or
            when received scheme doesn't match.

    Returns:
        Decorator to make callback to be set  up to require
        the `Authorization` header.

    Examples:
        ```python
        class MockEndpoint(WSGIEndpoint):

            @bearer_auth()
            def do_GET(self, token: str) -> None:
                # It is guaranteed that request headers include the
                # `Authorization` header at this point, and token is
                # the one extracted from the header.

                # Authenticate with any function working on your system.
                authenticate(token)

                # Do something...
        ```
    """
    def wrapper(callback: Callback_t) -> Callback_t:
        config = AuthSchemeConfig(callback)
        return config.set(AuthSchemes.bearer, err)

    return wrapper


@dataclasses.dataclass(eq=True, frozen=True)
class RequiredQueryInfo:

    query: str
    err_empty: t.Optional[ErrInfo] = None
    err_not_unique: t.Optional[ErrInfo] = None
    add_arg: bool = True


class RequiredQueryConfig(CallbackConfigBase):

    ATTR = _get_bamboo_attr("required_queries")

    def __init__(self, callback: Callback_t) -> None:
        super().__init__()

        if not hasattr(callback, self.ATTR):
            setattr(callback, self.ATTR, set())

        self._callback = callback
        self._registered: t.Set[RequiredQueryConfig] = getattr(callback, self.ATTR)

    def get(self) -> t.Tuple[RequiredHeaderInfo]:
        return tuple(self._registered)

    def set(self, info: RequiredQueryInfo) -> Callback_t:
        self._registered.add(info)

        if inspect.iscoroutinefunction(self._callback):
            func = self.decorate_asgi
        else:
            func = self.decorate_wsgi
        return func(self._callback, info)

    @staticmethod
    def decorate_wsgi(
        callback: Callback_WSGI_t,
        info: RequiredQueryInfo,
    ) -> Callback_WSGI_t:

        @functools.wraps(callback)
        def _callback(self: WSGIEndpoint, *args) -> None:
            val = self.get_queries(info.query)
            len_val = len(val)

            if len_val == 0:
                if info.err_empty:
                    raise info.err_empty
                else:
                    val = None
            elif len_val == 1:
                val = val[0]
            else:
                if info.err_not_unique:
                    raise info.err_not_unique

            if info.add_arg:
                callback(self, val, *args)
            else:
                callback(self, *args)

        errs = []
        if info.err_empty:
            errs.append(info.err_empty.__class__)
        if info.err_not_unique:
            errs.append(info.err_not_unique.__class__)
        if len(errs):
            _callback = may_occur(*errs)(_callback)

        return _callback

    @staticmethod
    def decorate_asgi(
        callback: Callback_ASGI_t,
        info: RequiredQueryInfo,
    ) -> Callback_ASGI_t:

        @functools.wraps(callback)
        async def _callback(self: ASGIHTTPEndpoint, *args) -> None:
            val = self.get_queries(info.query)
            len_val = len(val)

            if len_val == 0:
                if info.err_empty:
                    raise info.err_empty
                else:
                    val = None
            elif len_val == 1:
                val = val[0]
            else:
                if info.err_not_unique:
                    raise info.err_not_unique

            if info.add_arg:
                await callback(self, val, *args)
            else:
                await callback(self, *args)

        errs = []
        if info.err_empty:
            errs.append(info.err_empty.__class__)
        if info.err_not_unique:
            errs.append(info.err_not_unique.__class__)
        if len(errs):
            _callback = may_occur(*errs)(_callback)

        return _callback


def has_query_of(
    query: str,
    err_empty: t.Optional[ErrInfo] = None,
    err_not_unique: t.Optional[ErrInfo] = None,
    add_arg: bool = True,
) -> CallbackDecorator_t:
    """Set callback up to receive given query parameter from clients.
    """
    info = RequiredQueryInfo(query, err_empty, err_not_unique, add_arg)

    def wrapper(callback: Callback_t) -> Callback_t:
        config = RequiredQueryConfig(callback)
        return config.set(info)

    return wrapper


@dataclasses.dataclass(eq=True, frozen=True)
class SimpleAccessControlInfo:

    origins: t.Tuple[str] = ()
    allow_credentials: bool = False
    err_not_allowed: ErrInfo = DEFAULT_CORS_ERROR
    add_arg: bool = True


class SimpleAccessControlConfig(CallbackConfigBase):

    ATTR = _get_bamboo_attr("simple_access_control")

    def __init__(self, callback: Callback_t) -> None:
        super().__init__()

        self._callback = callback

    def get(self) -> t.Optional[SimpleAccessControlInfo]:
        return getattr(self._callback, self.ATTR, None)

    def set(self, info: SimpleAccessControlInfo) -> Callback_t:
        if hasattr(self._callback, self.ATTR):
            raise DuplicatedInfoError(
                "Decorating of multiple times is forbidden."
            )
        setattr(self._callback, self.ATTR, info)

        if inspect.iscoroutinefunction(self._callback):
            func = self.decorate_asgi
        else:
            func = self.decorate_wsgi
        return func(self._callback, info)

    @staticmethod
    def decorate_wsgi(
        callback: Callback_WSGI_t,
        info: SimpleAccessControlInfo,
    ) -> Callback_WSGI_t:
        origins = set(info.origins)

        @functools.wraps(callback)
        def _callback(self: WSGIEndpoint, *args) -> None:
            # Origin
            origin = self.get_header("Origin")
            if origin:
                if not len(origins):
                    self.add_header("Access-Control-Allow-Origin", "*")
                elif origin in origins:
                    self.add_header("Access-Control-Allow-Origin", origin)
                    self.add_header("Vary", "Origin")
                else:
                    raise info.err_not_allowed

            # Credentials
            if info.allow_credentials:
                self.add_header("Access-Control-Allow-Credentials", "true")

            if info.add_arg:
                callback(self, origin, *args)
            else:
                callback(self, *args)

        if info.err_not_allowed:
            _callback = may_occur(info.err_not_allowed.__class__)(_callback)
        return _callback

    @staticmethod
    def decorate_asgi(
        callback: Callback_ASGI_t,
        info: SimpleAccessControlInfo,
    ) -> Callback_ASGI_t:
        origins = set(info.origins)

        @functools.wraps(callback)
        async def _callback(self: ASGIHTTPEndpoint, *args) -> None:
            # Origin
            origin = self.get_header("Origin")
            if origin:
                if not len(origins):
                    self.add_header("Access-Control-Allow-Origin", "*")
                elif origin in origins:
                    self.add_header("Access-Control-Allow-Origin", origin)
                    self.add_header("Vary", "Origin")
                else:
                    raise info.err_not_allowed

            # Credentials
            if info.allow_credentials:
                self.add_header("Access-Control-Allow-Credentials", "true")

            if info.add_arg:
                await callback(self, origin, *args)
            else:
                await callback(self, *args)

        if info.err_not_allowed:
            _callback = may_occur(info.err_not_allowed.__class__)(_callback)
        return _callback


def allow_simple_access_control(
    *origins: str,
    allow_credentials: bool = False,
    err_not_allowed: ErrInfo = DEFAULT_CORS_ERROR,
    add_arg: bool = True,
) -> CallbackDecorator_t:
    info = SimpleAccessControlInfo(
        origins,
        allow_credentials,
        err_not_allowed,
        add_arg,
    )

    def wrapper(callback: Callback_t) -> Callback_t:
        config = SimpleAccessControlConfig(callback)
        return config.set(info)

    return wrapper


# NOTE
#   Each values should be lowercases.

_CORS_SAFELISTED_REQUEST_HEADERS = {
    "accept",
    "accept-language",
    "content-language",
    "content-type",
}


def _handle_cors_preflight(
    allow_methods: t.Union[str, t.Iterable[str]],
    allow_origins: t.Iterable[str] = (),
    allow_headers: t.Iterable[str] = (),
    expose_headers: t.Iterable[str] = (),
    max_age: t.Optional[int] = None,
    allow_credentials: bool = False,
    err_not_allowed_origin: ErrInfo = DEFAULT_CORS_ERROR,
    err_not_allowed_method: ErrInfo = DEFAULT_CORS_ERROR,
) -> t.Callable[
    [
        t.Union[ASGIEndpointBase, WSGIEndpoint],
        t.Optional[str],
        t.Optional[str],
        t.Optional[str],
    ],
    None
]:
    if isinstance(allow_methods, str):
        allow_methods = {allow_methods}
    allow_origins = set(allow_origins)
    allow_headers = set([header.lower() for header in allow_headers])
    allow_headers.update(_CORS_SAFELISTED_REQUEST_HEADERS)
    expose_headers = ", ".join(expose_headers)
    if max_age:
        max_age = str(max_age)

    def handle(
        self: t.Union[ASGIHTTPEndpoint, WSGIEndpoint],
        origin: t.Optional[str],
        req_method: t.Optional[str],
        req_headers: t.Optional[str],
    ) -> None:
        if origin is None and req_method is None:
            self.send_only_status(HTTPStatus.BAD_REQUEST)

        # Allow Origin
        if not len(allow_origins):
            self.add_header("Access-Control-Allow-Origin", "*")
        else:
            if origin in allow_origins:
                self.add_header("Access-Control-Allow-Origin", origin)
                self.add_header("Vary", "Origin")
            else:
                raise err_not_allowed_origin

        # Allow Methods
        if req_method is None:
            raise err_not_allowed_method
        if req_method in allow_methods:
            methods = ", ".join(allow_methods)
            self.add_header("Access-Control-Allow-Methods", methods)
        else:
            raise err_not_allowed_method

        # Allow Headers
        if req_headers:
            accepted_headers = ", ".join([
                header for header in re.split(",|, ", req_headers)
                if header in allow_headers
            ])
            if accepted_headers:
                self.add_header("Access-Control-Allow-Headers", accepted_headers)

        # Expose Headers
        if expose_headers:
            self.add_header("Access-Control-Expose-Headers", expose_headers)

        # Max Age
        if max_age:
            self.add_header("Access-Control-Max-Age", max_age)

        # Allow Credentials
        if allow_credentials:
            self.add_header("Access-Control-Allow-Credentials", "true")

        self.send_only_status(HTTPStatus.NO_CONTENT)

    return handle


def get_wsgi_preflight(
    allow_methods: t.Union[str, t.Iterable[str]],
    allow_origins: t.Iterable[str] = (),
    allow_headers: t.Iterable[str] = (),
    expose_headers: t.Iterable[str] = (),
    max_age: t.Optional[int] = None,
    allow_credentials: bool = False,
    err_not_allowed_origin: ErrInfo = DEFAULT_CORS_ERROR,
    err_not_allowed_method: ErrInfo = DEFAULT_CORS_ERROR,
) -> Callback_WSGI_t:
    callback = _handle_cors_preflight(
        allow_methods,
        allow_origins=allow_origins,
        allow_headers=allow_headers,
        expose_headers=expose_headers,
        max_age=max_age,
        allow_credentials=allow_credentials,
        err_not_allowed_origin=err_not_allowed_origin,
        err_not_allowed_method=err_not_allowed_method,
    )

    @has_header_of("Access-Control-Request-Headers")
    @has_header_of("Access-Control-Request-Method")
    @has_header_of("Origin")
    def do_OPTIONS(
        self: WSGIEndpoint,
        origin: t.Optional[str],
        req_method: t.Optional[str],
        req_headers: t.Optional[str],
    ) -> None:
        callback(self, origin, req_method, req_headers)

    return do_OPTIONS


def get_asgi_preflight(
    allow_methods: t.Union[str, t.Iterable[str]],
    allow_origins: t.Iterable[str] = (),
    allow_headers: t.Iterable[str] = (),
    expose_headers: t.Iterable[str] = (),
    max_age: t.Optional[int] = None,
    allow_credentials: bool = False,
    err_not_allowed_origin: ErrInfo = DEFAULT_CORS_ERROR,
    err_not_allowed_method: ErrInfo = DEFAULT_CORS_ERROR,
) -> Callback_ASGI_t:
    callback = _handle_cors_preflight(
        allow_methods,
        allow_origins=allow_origins,
        allow_headers=allow_headers,
        expose_headers=expose_headers,
        max_age=max_age,
        allow_credentials=allow_credentials,
        err_not_allowed_origin=err_not_allowed_origin,
        err_not_allowed_method=err_not_allowed_method,
    )

    @has_header_of("Access-Control-Request-Headers")
    @has_header_of("Access-Control-Request-Method")
    @has_header_of("Origin")
    async def do_OPTIONS(
        self: t.Type[ASGIHTTPEndpoint],
        origin: t.Optional[str],
        req_method: t.Optional[str],
        req_headers: t.Optional[str],
    ) -> None:
        callback(self, origin, req_method, req_headers)

    return do_OPTIONS


@dataclasses.dataclass(eq=True, frozen=True)
class PreFlightInfo:

    allow_methods: t.Tuple[str]
    allow_origins: t.Tuple[str] = ()
    allow_headers: t.Tuple[str] = ()
    expose_headers: t.Tuple[str] = ()
    max_age: t.Optional[int] = None
    allow_credentials: bool = False
    err_not_allowed_origin: ErrInfo = DEFAULT_CORS_ERROR
    err_not_allowed_method: ErrInfo = DEFAULT_CORS_ERROR
    add_arg: bool = True


class PreFlightConfig(HTTPEndpointConfigBase):

    ATTR = _get_bamboo_attr("preflight")

    def __init__(self, endpoint: t.Type[HTTPMixIn]) -> None:
        super().__init__()

        self._endpoint = endpoint

    def get(self) -> t.Optional[PreFlightInfo]:
        return getattr(self._endpoint, self.ATTR, None)

    def set(self, info: PreFlightInfo) -> t.Type[HTTPMixIn]:
        if hasattr(self._endpoint, self.ATTR):
            raise DuplicatedInfoError(
                "Decorating of multiple times is forbidden."
            )
        setattr(self._endpoint, self.ATTR, info)
        args = (
            info.allow_methods,
            info.allow_origins,
            info.allow_headers,
            info.expose_headers,
            info.max_age,
            info.allow_credentials,
            info.err_not_allowed_origin,
            info.err_not_allowed_method,
        )

        # Set other methods
        for name, res_method in self._endpoint._res_methods.copy().items():
            res_method = allow_simple_access_control(
                *info.allow_origins,
                allow_credentials=info.allow_credentials,
                err_not_allowed=info.err_not_allowed_origin,
                add_arg=info.add_arg,
            )(res_method)
            set_response_method(self._endpoint, name, res_method)

        # Set do_OPTIONS
        if issubclass(self._endpoint, WSGIEndpoint):
            res_method = get_wsgi_preflight(*args)
        elif issubclass(self._endpoint, ASGIHTTPEndpoint):
            res_method = get_asgi_preflight(*args)
        else:
            raise ValueError(
                f"Class {self._endpoint.__name__} is not avalidable."
            )
        set_response_method(self._endpoint, "OPTIONS", res_method)

        return self._endpoint


def add_preflight(
    allow_methods: t.Iterable[str],
    allow_origins: t.Iterable[str] = (),
    allow_headers: t.Iterable[str] = (),
    expose_headers: t.Iterable[str] = (),
    max_age: t.Optional[int] = None,
    allow_credentials: bool = False,
    err_not_allowed_origin: ErrInfo = DEFAULT_CORS_ERROR,
    err_not_allowed_method: ErrInfo = DEFAULT_CORS_ERROR,
    add_arg: bool = True,
) -> t.Callable[[HTTPMixIn], HTTPMixIn]:
    info = PreFlightInfo(
        tuple(allow_methods),
        allow_origins=tuple(allow_origins),
        allow_headers=tuple(allow_headers),
        expose_headers=tuple(expose_headers),
        max_age=max_age,
        allow_credentials=allow_credentials,
        err_not_allowed_origin=err_not_allowed_origin,
        err_not_allowed_method=err_not_allowed_method,
        add_arg=add_arg,
    )

    def wrapper(endpoint: HTTPMixIn) -> HTTPMixIn:
        config = PreFlightConfig(endpoint)
        return config.set(info)

    return wrapper


@dataclasses.dataclass(eq=True, frozen=True)
class CacheControlInfo:

    public: bool = False
    private: bool = False
    no_cache: bool = False
    no_store: bool = False
    max_age: t.Optional[int] = None
    s_maxage: t.Optional[int] = None
    max_stale: t.Optional[int] = None
    min_fresh: t.Optional[int] = None
    stale_while_revalidate: t.Optional[int] = None
    stale_if_error: t.Optional[int] = None
    must_revalidate: bool = False
    proxy_revalidate: bool = False
    immutable: bool = False
    no_transform: bool = False
    only_if_cached: bool = False


def _get_cache_control_value(info: CacheControlInfo) -> str:
    vals = []

    if info.public:
        vals.append("public")
    if info.private:
        vals.append("private")
    if info.no_cache:
        vals.append("no-cache")
    if info.no_store:
        vals.append("no-store")
    if info.max_age is not None:
        vals.append(f"max-age={info.max_age}")
    if info.s_maxage is not None:
        vals.append(f"s-maxage={info.s_maxage}")
    if info.max_stale is not None:
        vals.append(f"max-stale={info.max_stale}")
    if info.min_fresh is not None:
        vals.append(f"min-fresh={info.min_fresh}")
    if info.stale_while_revalidate is not None:
        vals.append(f"stale-while-revalidate={info.stale_while_revalidate}")
    if info.stale_if_error is not None:
        vals.append(f"stale-if-error={info.stale_if_error}")
    if info.must_revalidate:
        vals.append("must-revalidate")
    if info.proxy_revalidate:
        vals.append("proxy-revalidate")
    if info.immutable:
        vals.append("immutable")
    if info.no_transform:
        vals.append("no-transform")
    if info.only_if_cached:
        vals.append("only-if-cached")

    return ", ".join(vals)


class CacheControlConfig(CallbackConfigBase):

    ATTR = _get_bamboo_attr("cache_control")

    def __init__(self, callback: Callback_t) -> None:
        super().__init__()

        self._callback = callback

    def get(self) -> t.Optional[CacheControlInfo]:
        return getattr(self._callback, self.ATTR, None)

    def set(self, info: CacheControlInfo) -> Callback_t:
        if hasattr(self._callback, self.ATTR):
            raise DuplicatedInfoError(
                "Decorating of multiple times is forbidden."
            )
        setattr(self._callback, self.ATTR, info)

        if inspect.iscoroutinefunction(self._callback):
            func = self.decorate_asgi
        else:
            func = self.decorate_wsgi
        return func(self._callback, info)

    @staticmethod
    def decorate_wsgi(
        callback: Callback_WSGI_t,
        info: CacheControlInfo,
    ) -> Callback_WSGI_t:
        val = _get_cache_control_value(info)

        @functools.wraps(callback)
        def _callback(self: WSGIEndpoint, *args) -> None:
            self.add_header("Cache-Control", val)
            callback(self, *args)

        return _callback

    @staticmethod
    def decorate_asgi(
        callback: Callback_ASGI_t,
        info: CacheControlInfo,
    ) -> Callback_ASGI_t:
        val = _get_cache_control_value(info)

        @functools.wraps(callback)
        async def _callback(self: ASGIHTTPEndpoint, *args) -> None:
            self.add_header("Cache-Control", val)
            await callback(self, *args)

        return _callback


def set_cache_control(
    public: bool = False,
    private: bool = False,
    no_cache: bool = False,
    no_store: bool = False,
    max_age: t.Optional[int] = None,
    s_maxage: t.Optional[int] = None,
    max_stale: t.Optional[int] = None,
    min_fresh: t.Optional[int] = None,
    stale_while_revalidate: t.Optional[int] = None,
    stale_if_error: t.Optional[int] = None,
    must_revalidate: bool = False,
    proxy_revalidate: bool = False,
    immutable: bool = False,
    no_transform: bool = False,
    only_if_cached: bool = False,
) -> CallbackDecorator_t:
    info = CacheControlInfo(
        public=public,
        private=private,
        no_cache=no_cache,
        no_store=no_store,
        max_age=max_age,
        s_maxage=s_maxage,
        max_stale=max_stale,
        min_fresh=min_fresh,
        stale_while_revalidate=stale_while_revalidate,
        stale_if_error=stale_if_error,
        must_revalidate=must_revalidate,
        proxy_revalidate=proxy_revalidate,
        immutable=immutable,
        no_transform=no_transform,
        only_if_cached=only_if_cached,
    )

    def wrapper(callback: Callback_t) -> Callback_t:
        config = CacheControlConfig(callback)
        return config.set(info)

    return wrapper


SetCookieValue_t = t.Callable[
    [t.Union[ASGIHTTPEndpoint, WSGIEndpoint], str],
    None
]


def _handle_cookie(
    cookie_name: str,
    expires: t.Optional[str] = None,
    max_age: t.Optional[int] = None,
    domain: t.Optional[str] = None,
    path: t.Optional[str] = None,
    secure: bool = True,
    http_only: bool = True,
    samesite: t.Optional[str] = None,
) -> SetCookieValue_t:
    directives = []
    if expires is not None:
        directives.append(f"Expires={expires}")
    if max_age is not None:
        directives.append(f"Max-Age={max_age}")
    if domain is not None:
        directives.append(f"Domain={domain}")
    if path is not None:
        directives.append(f"Path={path}")
    if secure:
        directives.append("Secure")
    if http_only:
        directives.append("HttpOnly")
    if samesite is not None:
        directives.append(f"SameSite={samesite}")

    directives = "; ".join(directives)
    form = "{name}={value}; {others}"

    def set_cookie_value(
        self: t.Union[ASGIHTTPEndpoint, WSGIEndpoint],
        value: str,
    ) -> None:
        header_val = form.format(
            name=cookie_name,
            value=value,
            others=directives,
        )
        self.add_header("Set-Cookie", header_val)

    return set_cookie_value


@dataclasses.dataclass(eq=True, frozen=True)
class CookieInfo:

    cookie_name: str
    expires: t.Optional[str] = None
    max_age: t.Optional[int] = None
    domain: t.Optional[str] = None
    path: t.Optional[str] = None
    secure: bool = True
    http_only: bool = True
    samesite: t.Optional[str] = None


class CookieConfig(CallbackConfigBase):

    ATTR = _get_bamboo_attr("cookie")

    def __init__(self, callback: Callback_t) -> None:
        super().__init__()

        self._callback = callback

    def get(self) -> t.Optional[CookieInfo]:
        return getattr(self._callback, self.ATTR, None)

    def set(self, info: CookieInfo) -> Callback_t:
        if hasattr(self._callback, self.ATTR):
            raise DuplicatedInfoError(
                "Decorating of multiple times is forbidden."
            )
        setattr(self._callback, self.ATTR, info)

        if inspect.iscoroutinefunction(self._callback):
            func = self.decorate_asgi
        else:
            func = self.decorate_wsgi
        return func(self._callback, info)

    def decorate_wsgi(
        self,
        callback: Callback_WSGI_t,
        info: CookieInfo,
    ) -> Callback_WSGI_t:
        set_cookie_callback = _handle_cookie(
            info.cookie_name,
            expires=info.expires,
            max_age=info.max_age,
            domain=info.domain,
            path=info.path,
            secure=info.secure,
            http_only=info.http_only,
            samesite=info.samesite,
        )

        @functools.wraps(callback)
        def _callback(self: WSGIEndpoint, *args) -> None:
            callback(self, set_cookie_callback, *args)

        return _callback

    def decorate_asgi(
        self,
        callback: Callback_ASGI_t,
        info: CookieInfo,
    ) -> Callback_WSGI_t:
        set_cookie_callback = _handle_cookie(
            info.cookie_name,
            expires=info.expires,
            max_age=info.max_age,
            domain=info.domain,
            path=info.path,
            secure=info.secure,
            http_only=info.http_only,
            samesite=info.samesite,
        )

        @functools.wraps(callback)
        async def _callback(self: ASGIEndpointBase, *args) -> None:
            await callback(self, set_cookie_callback, *args)

        return _callback


def set_cookie(
    cookie_name: str,
    expires: t.Optional[str] = None,
    max_age: t.Optional[int] = None,
    domain: t.Optional[str] = None,
    path: t.Optional[str] = None,
    secure: bool = True,
    http_only: bool = True,
    samesite: t.Optional[str] = None,
) -> CallbackDecorator_t:
    info = CookieInfo(
        cookie_name,
        expires=expires,
        max_age=max_age,
        domain=domain,
        path=path,
        secure=secure,
        http_only=http_only,
        samesite=samesite,
    )

    def wrapper(callback: Callback_t) -> Callback_t:
        config = CookieConfig(callback)
        return config.set(info)

    return wrapper



from __future__ import annotations
from abc import ABCMeta, abstractmethod
import typing as t

from .asgi import (
    ASGIProtocols,
    ASGIRecv_t,
    ASGISend_t,
    ASGIWebSocketEvents,
    LifespanHandler_t,
    default_lifespan_handler,
    get_http_send_errinfo,
    get_http_sendbody,
    get_http_sendstart,
    get_websock_accept,
    get_websock_close,
    get_websock_recvmsg,
    get_websock_sendmsg,
)
from .endpoint import (
    ASGIEndpointBase,
    ASGIHTTPEndpoint,
    ASGIWebSocketEndpoint,
    EndpointBase,
    WSGIEndpoint,
)
from .error import DEFAULT_NOT_FOUND_ERROR, ErrInfo
from .io import BufferedConcatIterator
from .location import (
    Location_t,
    StaticLocation_t,
    Uri_t,
)
from .router import (
    Endpoint_t,
    Router,
    Uri2Endpoints_t,
)
from .sticky import _get_bamboo_attr


__all__ = []


ATTR_VERSION = _get_bamboo_attr("version")
Version_t = t.Tuple[int]


class VersionConfig:
    """Operator class for version of `Endpoint`.

    This class can be used to get and set version of `Endpoint` safely.
    """

    def __init__(self, endpoint: t.Type[EndpointBase]) -> None:
        """
        Args:
            endpoint: `Endpoint` whose version is to be manipulated
        """
        self._endpoint_class = endpoint

    def set(
        self,
        app: AppBase,
        version: t.Union[int, t.Tuple[int], None] = None,
    ) -> None:
        """Set version of `Endpoint`.

        Args:
            app: Application including the internal `Endpoint`.
            version: Version to be set.
            force: If forcing to set the `version`.

        Raises:
            ValueError: Raised if version of the `Endpoint` has already
                been set.
        """
        if not hasattr(self._endpoint_class, ATTR_VERSION):
            setattr(self._endpoint_class, ATTR_VERSION, {})

        registered = getattr(self._endpoint_class, ATTR_VERSION)

        # Format to fit the type Version_t
        if version is None:
            version = ()
        if isinstance(version, int):
            version = (version,)

        registered[app] = version

    def get(self, app: AppBase) -> t.Optional[Version_t]:
        """Retrieve version belonging to specified `app`.

        Args:
            app: Application including the internal `Endpoint`.

        Returns:
            Version set to `Endpoint`, if not set yet, then None.
        """
        if hasattr(self._endpoint_class, ATTR_VERSION):
            registered = getattr(self._endpoint_class, ATTR_VERSION)
            return registered.get(app)
        return None

    def get_all(self) -> t.List[t.Tuple[AppBase, Version_t]]:
        """Retrieve versions belonging to all `AppBase` objects.

        Returns:
            List of tuples of `AppBase` objects and their versions.
        """
        if hasattr(self._endpoint_class, ATTR_VERSION):
            registered = getattr(self._endpoint_class, ATTR_VERSION)
            return [(app, version) for app, version in registered.items()]
        return []


ATTR_PARCEL = _get_bamboo_attr("parcel")
Parcel_t = t.Tuple[t.Any, ...]


class ParcelConfig:
    """Operator class for parcel of `Endpoint`.

    This class can be used to get and set parcel of `Endpoint` safely.
    """

    def __init__(self, endpoint: t.Type[EndpointBase]) -> None:
        """
        Args:
            endpoint : `Endpoint` whose parcel is to be manipulated.
        """
        self._endpoint_class = endpoint

    def set(self, app: AppBase, parcel: Parcel_t) -> None:
        """Ser parcel of `Endpoint`.

        Args:
            app: Application including the internal `Endpoint`.
            parcel: Parcel to be set.
        """
        if not hasattr(self._endpoint_class, ATTR_PARCEL):
            setattr(self._endpoint_class, ATTR_PARCEL, {})

        registered = getattr(self._endpoint_class, ATTR_PARCEL)
        registered[app] = parcel

    def get(self, app: AppBase) -> Parcel_t:
        """Retrieve parcel belonging to specified `app`.

        Args:
            app: Application including the internal `Endpoint`

        Returns:
            Parcel set to `Endpoint`, if not set yet, then `None`.
        """
        if hasattr(self._endpoint_class, ATTR_PARCEL):
            registered = getattr(self._endpoint_class, ATTR_PARCEL)
            return registered.get(app)
        return ()

    def get_all(self) -> t.List[t.Tuple[AppBase, Parcel_t]]:
        """Retrieve parcels belonging to all `AppBase` objects.

        Returns:
            List of tuples of `AppBase` objects and their parcels.
        """
        if hasattr(self._endpoint_class, ATTR_PARCEL):
            registered = getattr(self._endpoint_class, ATTR_PARCEL)
            return [(app, parcel) for app, parcel in registered.items()]
        return []


class AppBase(t.Generic[Endpoint_t], metaclass=ABCMeta):
    """Base class of all application in Bamboo.

    Bamboo has two core concepts called application and endpoint, and
    this class implements basic behavior of the former, e.g. containing
    multiple endpoints, routing requests from URIs to endpoints and so on.

    Note:
        This class is an abstract class. Consider using its subclasses.

    Attributes:
        TAG_VERSION (str): Tag used when versions of `Endpoint`s are inserted
            in front of paths of URIs. If you want, you can override the
            value and set new favorite tag. By default, the tag is 'v'.
    """

    TAG_VERSION = "v"
    __avalidable_endpoints = (EndpointBase,)

    def __init__(
        self,
        error_404: ErrInfo = DEFAULT_NOT_FOUND_ERROR
    ) -> None:
        """
        Args:
            error_404: Error sending if a request to not registered URI or
                HTTP method comes.
        """
        self._router: Router[Endpoint_t] = Router()
        self._error_404 = error_404

    @abstractmethod
    def __call__(self, *args: t.Any, **kwds: t.Any) -> t.Any:
        pass

    def search_uris(self, endpoint: t.Type[Endpoint_t]) -> t.List[Uri_t]:
        """Retrieve all URI patterns of `Endpoint`.

        Note:
            This method uses `bamboo.Router.search_uris()` method inner.
            For more information, see the API document.

        Args:
            endpoint: `Endpoint` whose URIs to be searched.

        Returns:
            Result of searching.
        """
        return self._router.search_uris(endpoint)

    def validate(
        self,
        uri: str,
    ) -> t.Tuple[t.Tuple[str, ...], t.Optional[t.Type[Endpoint_t]]]:
        """Validate specified `uri` and retrieve corresponding `Endpoint` class.

        Note:
            This method uses `bamboo.Router.validate()` method inner.
            For more information, see the API document.

        Args:
            uri: Path of URI to be validated.

        Returns:
            Pair of values of flexible locations and `Endpoint` class if specified
            `uri` is valid, otherwise, pari of empty tuple and None.
        """
        return self._router.validate(uri)

    def route(
        self,
        *locs: Location_t,
        version: t.Union[int, t.Tuple[int], None] = None
    ) -> t.Callable[[t.Type[Endpoint_t]], t.Type[Endpoint_t]]:
        """Register combination of URI and `Endpoint` for routing.

        Args:
            *locs: Locations of path of the URI bound with the `Endpoint`.
            version : Version of the `Endpoint`.

        Returns:
            Decorator to add combination of URI and `Endpoint`.

        Examples:
            ```python
            app = App()

            # Set path of URI as `test/data/image` and the version as 1
            @app.route("test", "data", "image", version=1)
            class MockEndpoint(Endpoint):

                def do_GET(self) -> None:
                    # Do something...
            ```
        """
        def register_endpoint(
            endpoint: t.Type[Endpoint_t],
        ) -> t.Type[Endpoint_t]:
            if not issubclass(endpoint, self.__avalidable_endpoints):
                raise TypeError(
                    f"{endpoint.__name__} is not avalidable in "
                    f"the {self.__class__.__name__}."
                )

            # version setting
            ver_config = VersionConfig(endpoint)
            ver_config.set(self, version)

            # router setting
            _version = ver_config.get(self)
            assert _version is not None
            if len(_version):
                _version = tuple(f"{self.TAG_VERSION}{v}" for v in _version)
            self._router.register(locs, endpoint, version=_version)

            return endpoint
        return register_endpoint

    def set_parcel(self, endpoint: t.Type[Endpoint_t], *parcel: t.Any) -> None:
        """Set parcel to an endpoint.

        This method enables to give objects to `Endpoint` objects
        dynamically. A parcel is a set of objects delivered via the method,
        and the `Endpoint` object of its destination receives it at
        `EndpointBase.setup()` method.

        Note:
            For more information about the `bamboo.EndpointBase.setup()`,
            see the API document.

        Args:
            endpoint: `Endpoint` the `parcel` to be set.
            *parcel: Pacel to be given to the `Endpoint`.
        """
        parcel_config = ParcelConfig(endpoint)
        parcel_config.set(self, parcel)

    @property
    def tree(self) -> Uri2Endpoints_t:
        """Tree of the application's endpoints.
        """
        return self._router.uri2endpoint.copy()

    def graft(
        self,
        *apps: AppBase,
        onto: t.Tuple[StaticLocation_t, ...] = ()
    ) -> None:
        """Graft other applications as branches of the application's tree.

        Args:
            *apps: Branch applications.
            onto: Root path of the branches.
        """
        for app in apps:
            for locs, endpoint in app._router._raw_uri2endpoint.items():
                locs = onto + locs

                ver_config = VersionConfig(endpoint)
                version = ver_config.get(app)
                ver_config.set(self, version=version)
                if len(version):
                    version = tuple(f"{self.TAG_VERSION}{v}" for v in version)

                self._router.register(locs, endpoint, version=version)


WSGIStartRespoint_t = t.Callable[
    [str, t.Iterable[t.Tuple[str, str]]],
    None
]


class WSGIApp(AppBase):
    """Application compliant with the WSGI.

    This class is a subclass of `AppBase` calss and implements the callbable
    compliant with the WSGI.

    Note:
        This class can be used only for WSGI server. If you want to use
        any ASGI servers, consider using `ASGIHTTPApp`.

        This class can also route only `WSGIEndpoint`s. If you want to
        another type of endpoint, consider implementation class of its
        corresponding application.
    """

    __avalidable_endpoints = (WSGIEndpoint,)

    def __call__(
        self,
        environ: t.Dict[str, t.Any],
        start_response: WSGIStartRespoint_t,
    ) -> t.List[bytes]:
        method = environ.get("REQUEST_METHOD").upper()
        path = environ.get("PATH_INFO")

        flexible_locs, endpoint_class = self.validate(path)
        if endpoint_class is None:
            return self.send_404(start_response)

        parcel_config = ParcelConfig(endpoint_class)
        parcel = parcel_config.get(self)

        pre_callback = endpoint_class._get_pre_response_method(method)
        callback = endpoint_class._get_response_method(method)
        if callback is None:
            return self.send_404(start_response)

        endpoint = endpoint_class(self, environ, flexible_locs, *parcel)
        # NOTE
        #   Subclasses of the ErrInfo must be raised in pre-response
        #   methods or response methods. Otherwise, the errors behave
        #   like ordinary exception objects.
        try:
            if pre_callback:
                pre_callback(endpoint)
            callback(endpoint)
        except ErrInfo as e:
            status, headers, body = e.get_all_form()
            # NOTE
            #   Other exception not inheriting the ErrInfo class
            #   are not to be catched here.

            for header_name, header_value in endpoint._res_headers:
                if e.should_inherit_header(header_name):
                    headers.append((header_name, header_value))
        else:
            status = endpoint._res_status
            headers = endpoint._res_headers
            body = endpoint._res_body

        start_response(status.wsgi, headers)
        return body

    def send_404(
        self,
        start_response: WSGIStartRespoint_t,
    ) -> BufferedConcatIterator:
        """Send `404` error code, i.e. `Resource Not Found` error.

        Args:
            start_response: `start_response` callable given from
                the WSGI application.

        Returns:
            Response body of the error.
        """
        status, headers, res_body = self._error_404.get_all_form()
        start_response(status.wsgi, headers)
        return res_body

    def search_uris(self, endpoint: t.Type[WSGIEndpoint]) -> t.List[Uri_t]:
        return super().search_uris(endpoint)

    def validate(
        self,
        uri: str
    ) -> t.Tuple[t.Tuple[str, ...], t.Optional[t.Type[WSGIEndpoint]]]:
        return super().validate(uri)

    def route(
        self,
        *locs: Location_t,
        version: t.Union[int, t.Tuple[int], None] = None
    ) -> t.Callable[[t.Type[WSGIEndpoint]], t.Type[WSGIEndpoint]]:
        return super().route(*locs, version=version)

    def set_parcel(
        self,
        endpoint: t.Type[WSGIEndpoint],
        *parcel: t.Any
    ) -> None:
        return super().set_parcel(endpoint, *parcel)


class ASGIApp(AppBase):
    """Application compliant with the ASGI.

    This class is a subclass of `AppBase` calss and implements the callbable
    compliant with the ASGI.

    Note:
        This class can be used only for ASGI server. If you want to use
        any WSGI servers, consider using `WSGIApp`.

        This class can also route only `ASGIHTTPEndpoint`s and
        `ASGIWebSocketEndpoint`s. If you want to another type of endpoint,
        consider implementation class of its corresponding application.
    """

    __avalidable_endpoints = (ASGIHTTPEndpoint, ASGIWebSocketEndpoint)

    def __init__(
        self,
        error_404: ErrInfo = DEFAULT_NOT_FOUND_ERROR,
        lifespan_handler: LifespanHandler_t = default_lifespan_handler,
    ) -> None:
        """
        Args:
            error_404: Error sending if a request to not registered URI or
                HTTP method comes.
            lifespan_handler: A callback with the type `LifespanHandler_t`.
        """
        super().__init__(error_404=error_404)

        self._lifespan_handler = lifespan_handler

    async def __call__(
        self,
        scope: t.Dict[str, t.Any],
        recv: ASGIRecv_t,
        send: ASGISend_t,
    ) -> None:
        """Handle requests compliant with the ASGI.

        Args:
            scope: Connection scope in the ASGI.
            recv: Awaitable callable to receive new event data.
            send: Awaitable callable to send new event data.
        """
        typ = scope.get("type")
        if typ == ASGIProtocols.http:
            await self.handle_http(scope, recv, send)
        elif typ == ASGIProtocols.websocket:
            await self.handle_websocket(scope, recv, send)
        elif typ == ASGIProtocols.lifespan:
            await self.handle_lifespan(scope, recv, send)
        else:
            raise NotImplementedError

    async def handle_http(
        self,
        scope: t.Dict[str, t.Any],
        recv: ASGIRecv_t,
        send: ASGISend_t,
    ) -> None:
        """Handle requests with the HTTP protocol.

        This method only handles requests with the HTTP and is used in
        the `__call__()` method for the ASGI application. Therefore one
        doesn't have to call this method if one just runs the application.

        Args:
            scope: Connection scope in the ASGI.
            recv: Awaitable callable to receive new event data.
            send: Awaitable callable to send new event data.
        """
        method = scope.get("method")
        path = scope.get("path")
        sendstart = get_http_sendstart(send)
        sendbody = get_http_sendbody(send)
        send_errinfo = get_http_send_errinfo(send)

        flexible_locs, endpoint_class = self.validate(path)
        if endpoint_class is None:
            await send_errinfo(self._error_404, ())
            return

        parcel_config = ParcelConfig(endpoint_class)
        parcel = parcel_config.get(self)

        pre_callback = endpoint_class._get_pre_response_method(method)
        callback = endpoint_class._get_response_method(method)
        if callback is None:
            await send_errinfo(self._error_404, ())
            return

        endpoint = endpoint_class(self, scope, recv, flexible_locs, *parcel)
        # NOTE
        #   Subclasses of the ErrInfo must be raised in pre-response
        #   methods or response methods. Otherwise, the errors behave
        #   like ordinary exception objects.
        try:
            if pre_callback:
                await pre_callback(endpoint)
            await callback(endpoint)
        except ErrInfo as e:
            await send_errinfo(e, endpoint._res_headers)
            return

            # NOTE
            #   Other exceptions not inheriting the ErrInfo class
            #   are not to be catched here.
        else:
            status = endpoint._res_status
            headers = endpoint._res_headers
            body = endpoint._res_body

        await sendstart(status, headers)
        await sendbody(body)

    async def handle_websocket(
        self,
        scope: t.Dict[str, t.Any],
        recv: ASGIRecv_t,
        send: ASGISend_t,
    ) -> None:
        """Handle requests with the WebSocket protocol.

        This method only handles requests with the WebSocket and is used in
        the `__call__()` method for the ASGI application. Therefore one
        doesn't have to call this method if one just runs the application.

        Args:
            scope: Connection scope in the ASGI.
            recv: Awaitable callable to receive new event data.
            send: Awaitable callable to send new event data.
        """
        path = scope.get("path")

        flexible_locs, endpoint_class = self.validate(path)
        if endpoint_class is None:
            await self.send_404(send)
            return

        parcel_config = ParcelConfig(endpoint_class)
        parcel = parcel_config.get(self)
        endpoint = endpoint_class(self, scope, flexible_locs, *parcel)

        # Establish connection
        msg = await recv()
        assert msg["type"] == ASGIWebSocketEvents.connect
        accept = get_websock_accept(send)
        await endpoint.do_ACCEPT(accept)

        # Main communications
        recvmsg = get_websock_recvmsg(recv)
        sendmsg = get_websock_sendmsg(send)
        close = get_websock_close(send)
        await endpoint.do_COMMUNICATE(recvmsg, sendmsg, close)

    async def handle_lifespan(
        self,
        scope: t.Dict[str, t.Any],
        recv: ASGIRecv_t,
        send: ASGISend_t,
    ) -> None:
        """Handle requests with the Lifespan protocol.

        This method only handles requests with the Lifespan and is used in
        the `__call__()` method for the ASGI application. Therefore one
        doesn't have to call this method if one just runs the application.

        Args:
            scope: Connection scope in the ASGI.
            recv: Awaitable callable to receive new event data.
            send: Awaitable callable to send new event data.
        """
        await self._lifespan_handler(scope, recv, send)

    def apply_lifespan_handler(
        self,
        lifespan_handler: LifespanHandler_t,
    ) -> None:
        """Apply a handler for the Lifespan protocol.

        This method sets a callback with the type `LifespanHandler_t` to
        handle requests with the Lifespan protocol. This method should be
        called only if one intends to set one's custom callback.

        Args:
            lifespan_handler: A callback with the type `LifespanHandler_t`.
        """
        self._lifespan_handler = lifespan_handler

    def search_uris(
        self,
        endpoint: t.Type[ASGIHTTPEndpoint],
    ) -> t.List[Uri_t]:
        return super().search_uris(endpoint)

    def validate(
        self,
        uri: str,
    ) -> t.Tuple[t.Tuple[str, ...], t.Optional[t.Type[ASGIEndpointBase]]]:
        return super().validate(uri)

    def route(
        self,
        *locs: Location_t,
        version: t.Union[int, t.Tuple[int], None] = None,
    ) -> t.Callable[[t.Type[ASGIEndpointBase]], t.Type[ASGIEndpointBase]]:
        return super().route(*locs, version=version)

    def set_parcel(
        self,
        endpoint: t.Type[ASGIEndpointBase],
        *parcel: t.Any,
    ) -> None:
        return super().set_parcel(endpoint, *parcel)



import base64
import typing as t


__all__ = [
    "decode2binary",
    "encode_binary",
    "encode_base64_string",
    "unparse_qs",
]


def encode_binary(data: t.Union[bytes, bytearray]) -> str:
    return base64.b64encode(data).decode()


def encode_base64_string(data: str) -> str:
    return encode_binary(data.encode())


def decode2binary(data: str) -> bytes:
    return base64.b64decode(data.encode())


def unparse_qs(query: t.Dict[str, t.List[str]]) -> str:
    params = []
    for key, vals in query.items():
        repr_vals = ",".join(vals)
        params.append("=".join((key, repr_vals)))
    return "&".join(params)



from __future__ import annotations
from abc import ABCMeta, abstractmethod
import typing as t

from ..http import (
    ContentType,
    ContentTypeHolder,
    DEFAULT_CONTENT_TYPE_PLAIN,
)
from ..util.deco import class_property


class ApiValidationFailedError(Exception):
    """Raised if type validation of data failes."""
    pass


class ApiData(ContentTypeHolder, metaclass=ABCMeta):
    """Base class to describe input/output data format on Endpoints as APIs.

    Subclasses of this class can be used an argument of `data_format`
    decorator for callbacks on Endpoints. If input/output parameters of
    the `data_format` are specified, the decorator validates if raw data,
    typically a `bytes` object, has expected data format.

    Subclasses of this class can define their own data formats in `__init__`
    methods. Developers should implement `__init__` methods of the subclasses
    such that each objects has data with expected formats by implementors.

    Note:
        This class is an abstract class. So, don't initilize it and specify it
        as an argument of the `data_format` decorator.

        Subclasses of this class should validate if raw data given to
        `__init__` methods has expected formats. If the validation failes,
        the classes MUST raise `ValidataionFailedError` to announce the
        failure to the `data_format` decorator.
    """

    @classmethod
    @abstractmethod
    def __validate__(cls, raw: bytes, content_type: ContentType) -> ApiData:
        """
        Args:
            raw : Raw data to be validated.
            content_type : Values of `Content-Type` header.
        """
        pass

    @abstractmethod
    def __extract__(self) -> t.Union[bytes, t.Iterable[bytes]]:
        pass


class BinaryApiData(ApiData):
    """API data with no format.

    This class can be used to describe raw data with no data format. So,
    any received data from clients is acceptable on the class.

    Examples:
        ```python
        class MockEndpoint(Endpoint):

            @data_format(input=BinaryApiData, output=None)
            def do_GET(self, rec_body: BinaryApiData) -> None:
                # get raw data of request body
                raw_data = rec_body.raw
        ```
    """

    def __init__(self, data: bytes) -> None:
        """
        Args:
            data: Binary data.
        """
        self._data = data

    @classmethod
    def __validate__(cls, raw: bytes, content_type: ContentType) -> BinaryApiData:
        """
        Args:
            raw: Raw data to be validated.
            content_type: Values of `Content-Type` header.

        Returns:
            The BinaryApiData object validated succcessfully.

        Note:
            In objects of this class, `content_type` is not used even if any
            `content_type` is specified.
        """
        if not isinstance(raw, bytes):
            raise ApiValidationFailedError(f"'raw' must be a 'bytes'.")
        return cls(raw)

    @property
    def raw(self) -> bytes:
        """Raw data of input binary."""
        return self._data

    @class_property
    def __content_type__(cls) -> ContentType:
        """Content type with `text/plain`.
        """
        return DEFAULT_CONTENT_TYPE_PLAIN

    def __extract__(self) -> bytes:
        return self.raw



from .app import (
    AppBase,
    ASGIApp,
    ParcelConfig,
    VersionConfig,
    WSGIApp,
)
from .asgi import (
    ASGIHTTPEvents,
    ASGILifespanEvenets,
    ASGIRecv_t,
    ASGISend_t,
    ASGIWebSocketEvents,
    LifespanHandler_t,
    WebSocketAccept_t,
    WebSocketClose_t,
    WebSocketDisconnectedError,
    WebSocketError,
    WebSocketRecvMsg_t,
    WebSocketSendMsg_t,
    default_lifespan_handler,
)
from .http import (
    AuthSchemes,
    ContentType,
    ContentTypeHolder,
    DEFAULT_CONTENT_TYPE_JSON,
    DEFAULT_CONTENT_TYPE_PLAIN,
    HTTPMethods,
    HTTPStatus,
    MediaTypes,
)
from .endpoint import (
    ASGIEndpointBase,
    ASGIHTTPEndpoint,
    ASGIWebSocketEndpoint,
    EndpointBase,
    HTTPMixIn,
    StatusCodeAlreadySetError,
    WSGIEndpoint,
    WSGIEndpointBase,
)
from .error import (
    ApiErrInfo,
    DEFAULT_BASIC_AUTH_HEADER_NOT_FOUND_ERROR,
    DEFAULT_BEARER_AUTH_HEADER_NOT_FOUND_ERROR,
    DEFAULT_HEADER_NOT_FOUND_ERROR,
    DEFUALT_INCORRECT_DATA_FORMAT_ERROR,
    DEFAULT_NOT_APPLICABLE_IP_ERROR,
    DEFAULT_NOT_FOUND_ERROR,
    DEFAULT_QUERY_PARAM_NOT_FOUND_ERROR,
    ErrInfo,
    get_auth_realm,
    get_default_auth_realm,
)
from .io import (
    BufferedBinaryIterator,
    BufferedConcatIterator,
    BufferedFileIterator,
    BufferedIteratorWrapper,
    BufferedStreamIterator,
    ResponseBodyIteratorBase,
)
from .location import (
    AnyStringLocation,
    AsciiDigitLocation,
    FlexibleLocation,
    is_duplicated_uri,
    is_flexible_uri,
)
from .router import (
    DuplicatedUriRegisteredError,
    Router,
)
from .test import (
    WSGIServerForm,
    WSGITestExecutor,
)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bamboo', 'bamboo.api', 'bamboo.request', 'bamboo.sticky', 'bamboo.util']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'bamboo-core',
    'version': '0.10.2',
    'description': 'General purpose server framework in Python',
    'long_description': '# Bamboo\n\n[![Bamboo](docs/res/bamboo.png)](https://jjj999.github.io/bamboo/)\n[![PyPI version](https://badge.fury.io/py/bamboo-core.svg)](http://badge.fury.io/py/bamboo-core)\n[![License](https://img.shields.io/github/license/mashape/apistatus.svg)](https://pypi.python.org/pypi/bamboo-core/)\n[![](https://img.shields.io/badge/docs-stable-blue.svg)](https://jjj999.github.io/bamboo)\n\n## Supported Interfaces\n\n- WSGI\n- ASGI v3.0 (HTTP, WebSocket and Lifespan)\n\n## Installling\n\n* Python: >= 3.7\n\n```\n$ python -m pip install bamboo-core\n```\n\n## [Usage](https://jjj999.github.io/bamboo/tutorials/concept/)\n\n以下は簡単な実装例です．\n\n```python\nfrom bamboo import WSGIApp, WSGIEndpoint, WSGITestExecutor\n\napp = WSGIApp()\n\n@app.route("hello")\nclass MockEndpoint(WSGIEndpoint):\n\n    def do_GET(self) -> None:\n        self.send_body(b"Hello, World!")\n\nif __name__ == "__main__":\n    WSGITestExecutor.debug(app)\n```\n\n上記スクリプトを実行後，ブラウザで http://localhost:8000/hello にアクセスするとレスポンスを確認できます．\n\n## API documentation\n\nAPI ドキュメントは[こちら](https://jjj999.github.io/bamboo/api/bamboo/pkg/)．\n\n## Examples\n\n### [upsidedown](https://github.com/jjj999/bamboo/tree/master/examples/upsidedown)\n\nリクエストされた文字列を逆順に反転させて返すアプリケーションです．\n\n### [image_traffic](https://github.com/jjj999/bamboo/tree/master/examples/image_traffic)\n\nアクセスに対して静的な画像を返すアプリケーションです．\n\n### [tweets](https://github.com/jjj999/bamboo/tree/master/examples/tweets)\n\nCLI ベースの簡易的な Twitter のような投稿アプリです．認証機能は実装されていません．\n',
    'author': 'jjj999',
    'author_email': 'jjj999to@gmail.com',
    'maintainer': 'jjj999',
    'maintainer_email': 'jjj999to@gmail.com',
    'url': 'https://jjj999.github.io/bamboo',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations
import typing as t
import urllib.parse

from .base import ApiData, ApiValidationFailedError
from ..http import ContentType, MediaTypes
from ..util.deco import cached_property, class_property


TYPES_ARGS = (str,)


def _check_annotation(typ: t.Type) -> None:
    if typ not in TYPES_ARGS:
        raise TypeError(
            f"{typ.__name__} is unacceptable as a type "
            "of attributes of FormApiData."
        )


def _has_valid_annotations(apiclass: t.Type[FormApiData]) -> None:
    for typ in t.get_type_hints(apiclass).values():
        _check_annotation(typ)


def _build_dict(
    apiclass: t.Type[FormApiData],
    **data: str,
) -> FormApiData:
    instance = apiclass.__new__(apiclass)
    annotations = t.get_type_hints(apiclass)

    for key in annotations.keys():
        if key not in data:
            raise ApiValidationFailedError(
                f"Raw data has no key of '{key}'."
            )
        if type(data[key]) not in TYPES_ARGS:
            raise ApiValidationFailedError(
                "Invalid type was detected. "
                f"Expected {apiclass.__name__}.{key} was a str."
            )
        setattr(instance, key, data[key])
    return instance


def _build_form_api(
    apiclass: t.Type[FormApiData],
    data: str
) -> FormApiData:
    parsing = {}
    for key, val in urllib.parse.parse_qs(data).items():
        if len(val) > 1:
            raise ApiValidationFailedError(
                f"Duplicated keys '{key}' were detected."
            )
        parsing[key] = val[0]
    return _build_dict(apiclass, **parsing)


class FormApiData(ApiData):
    """API data with `x-www-form-urlencoded`

    This class can be used to describe data with `x-www-form-urlencoded`
    format. This class should be inheritted and its several
    class-attributes should be defined in the subclass. Developer must
    define type hints of the class-attributes, which are used to validate if
    raw data has format the type hints define. In this class, the type hints
    must be only `str`. Otherwise, `TypeError` will be raised.

    Examples:
        - Defining subclass of this class
        ```python
        class UserCredentials(FormApiData):
            user_id: str
            password: str
        ```

        - Validating received data
        ```python
        class MockEndpoint(Endpoint):

            @data_format(input=UserCredentials, output=None)
            def do_GET(self, rec_body: UserCredentials) -> None:
                # Do something...

                # Example
                authenticate(rec_body.user_id, rec_body.password)
        ```
    """

    def __init_subclass__(cls) -> None:
        _has_valid_annotations(cls)

    def __init__(self, **data: str) -> None:
        """
        Args:
            raw: Raw data to be validated.
            content_type: Values of `Content-Type` header.
        """
        mapped = _build_dict(self.__class__, **data)
        self.__dict__.update(mapped.__dict__)

    @classmethod
    def __validate__(
        cls,
        raw: bytes,
        content_type: ContentType,
    ) -> FormApiData:
        """
        Args:
            raw: Raw data to be validated.
            content_type: Values of `Content-Type` header.

        Returns:
            The FormApiData object validated successfully.
        """
        if content_type.charset is None:
            content_type.charset = "UTF-8"

        if not cls.verify_content_type(content_type):
            raise ApiValidationFailedError(
                "Media type of 'Content-Type' header is not "
                f"{MediaTypes.x_www_form_urlencoded}, "
                f"but {content_type.media_type}."
            )

        try:
            raw = raw.decode(encoding=content_type.charset)
        except UnicodeDecodeError as e:
            raise ApiValidationFailedError(
                "Decoding raw data failed. The encoding was expected "
                f"{content_type.charset}, but not corresponded."
            ) from e
        return _build_form_api(cls, raw)

    def __extract__(self) -> bytes:
        encoding = self.__content_type__.charset
        if encoding is None:
            encoding = "utf-8"
        return self.string.encode(encoding=encoding)

    @cached_property
    def dict(self) -> t.Dict[str, t.Any]:
        keys = t.get_type_hints(self.__class__).keys()
        return {key: getattr(self, key) for key in keys}

    @cached_property
    def string(self) -> str:
        return "&".join([f"{key}={val}" for key, val in self.dict.items()])

    @class_property
    def __content_type__(cls) -> ContentType:
        """Content type with `application/x-www-form-urlencoded`.
        """
        return ContentType(MediaTypes.x_www_form_urlencoded)
